# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum

from PyQt5.QtCore import QT_TRANSLATE_NOOP
from PyQt5.QtWidgets import QComboBox

from lisp.cues.cue import Cue, CueAction
from lisp.ui.ui_utils import translate

QT_TRANSLATE_NOOP('CueAction', 'Default')
QT_TRANSLATE_NOOP('CueAction', 'FadeInStart')
QT_TRANSLATE_NOOP('CueAction', 'FadeOutStop')
QT_TRANSLATE_NOOP('CueAction', 'FadeOutPause')
QT_TRANSLATE_NOOP('CueAction', 'Start')
QT_TRANSLATE_NOOP('CueAction', 'Stop')
QT_TRANSLATE_NOOP('CueAction', 'Pause')


class CueActionComboBox(QComboBox):
    class Mode(Enum):
        Action = 0
        Value = 1
        Name = 2

    def __init__(self, cue_class, mode=Mode.Action, **kwargs):
        super().__init__(**kwargs)
        self.mode = mode

        if issubclass(cue_class, Cue):
            for action in cue_class.CueActions:
                if mode is CueActionComboBox.Mode.Value:
                    value = action.value
                elif mode is CueActionComboBox.Mode.Name:
                    value = action.name
                else:
                    value = action

                self.addItem(translate('CueAction', action.name), value)

    def currentAction(self):
        return self.currentData()

    def setCurrentAction(self, action):
        if self.mode is CueActionComboBox.Mode.Value:
            name = CueAction(action).name
        elif self.mode is CueActionComboBox.Mode.Action:
            name = action.name
        else:
            name = action

        self.setCurrentText(translate('CueAction', name))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import subprocess

from PyQt5.QtCore import Qt, QT_TRANSLATE_NOOP
from PyQt5.QtWidgets import QVBoxLayout, QGroupBox, QLineEdit, QCheckBox

from lisp.core.decorators import async
from lisp.core.has_properties import Property
from lisp.cues.cue import Cue, CueState, CueAction
from lisp.ui.settings.cue_settings import CueSettingsRegistry
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class CommandCue(Cue):
    """Cue able to execute system commands.

    Implemented using :class:`subprocess.Popen` with *shell=True*
    """

    Name = QT_TRANSLATE_NOOP('CueName', 'Command Cue')
    CueActions = (CueAction.Default, CueAction.Start, CueAction.Stop)

    command = Property(default='')
    no_output = Property(default=True)
    no_error = Property(default=True)
    kill = Property(default=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = translate('CueName', self.Name)

        self.__process = None
        self.__stopped = False

    def __start__(self, fade=False):
        self.__exec_command()
        return True

    @async
    def __exec_command(self):
        if not self.command.strip():
            return

        # If no_output is True, discard all the outputs
        std = subprocess.DEVNULL if self.no_output else None
        # Execute the command
        self.__process = subprocess.Popen(self.command, shell=True,
                                          stdout=std, stderr=std)
        rcode = self.__process.wait()

        if rcode == 0 or rcode == -9 or self.no_error:
            # If terminate normally, killed or in no-error mode
            if not self.__stopped:
                self._ended()

            self.__process = None
            self.__stopped = False
        elif not self.no_error:
            # If an error occurs and not in no-error mode
            self._error(
                translate('CommandCue', 'Process ended with an error status.'),
                translate('CommandCue', 'Exit code: ') + str(rcode))

    def __stop__(self, fade=False):
        if self.__process is not None:
            self.__stopped = True
            if self.kill:
                self.__process.kill()
            else:
                self.__process.terminate()

        return True

    __interrupt__ = __stop__


class CommandCueSettings(SettingsPage):
    Name = QT_TRANSLATE_NOOP('SettingsPageName', 'Command')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.group = QGroupBox(self)
        self.group.setLayout(QVBoxLayout(self.group))
        self.layout().addWidget(self.group)

        self.commandLineEdit = QLineEdit(self.group)
        self.group.layout().addWidget(self.commandLineEdit)

        self.noOutputCheckBox = QCheckBox(self)
        self.layout().addWidget(self.noOutputCheckBox)

        self.noErrorCheckBox = QCheckBox(self)
        self.layout().addWidget(self.noErrorCheckBox)

        self.killCheckBox = QCheckBox(self)
        self.layout().addWidget(self.killCheckBox)

        self.retranslateUi()

    def retranslateUi(self):
        self.group.setTitle(translate('CommandCue', 'Command'))
        self.commandLineEdit.setPlaceholderText(
            translate('CommandCue', 'Command to execute, as in a shell'))
        self.noOutputCheckBox.setText(
            translate('CommandCue', 'Discard command output'))
        self.noErrorCheckBox.setText(
            translate('CommandCue', 'Ignore command errors'))
        self.killCheckBox.setText(
            translate('CommandCue', 'Kill instead of terminate'))

    def enable_check(self, enabled):
        self.group.setCheckable(enabled)
        self.group.setChecked(False)

        self.noOutputCheckBox.setTristate(enabled)
        if enabled:
            self.noOutputCheckBox.setCheckState(Qt.PartiallyChecked)

        self.noErrorCheckBox.setTristate(enabled)
        if enabled:
            self.killCheckBox.setCheckState(Qt.PartiallyChecked)

        self.killCheckBox.setTristate(enabled)
        if enabled:
            self.killCheckBox.setCheckState(Qt.PartiallyChecked)

    def load_settings(self, settings):
        self.commandLineEdit.setText(settings.get('command', ''))
        self.noOutputCheckBox.setChecked(settings.get('no_output', True))
        self.noErrorCheckBox.setChecked(settings.get('no_error', True))
        self.killCheckBox.setChecked(settings.get('kill', False))

    def get_settings(self):
        settings = {}

        if not (self.group.isCheckable() and not self.group.isChecked()):
            if self.commandLineEdit.text().strip():
                settings['command'] = self.commandLineEdit.text()
        if self.noOutputCheckBox.checkState() != Qt.PartiallyChecked:
            settings['no_output'] = self.noOutputCheckBox.isChecked()
        if self.noErrorCheckBox.checkState() != Qt.PartiallyChecked:
            settings['no_error'] = self.noErrorCheckBox.isChecked()
        if self.killCheckBox.checkState() != Qt.PartiallyChecked:
            settings['kill'] = self.killCheckBox.isChecked()

        return settings


CueSettingsRegistry().add_item(CommandCueSettings, CommandCue)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum

from PyQt5.QtCore import QT_TRANSLATE_NOOP
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QComboBox, QStyledItemDelegate

from lisp.ui.ui_utils import translate

QT_TRANSLATE_NOOP('Fade', 'Linear')
QT_TRANSLATE_NOOP('Fade', 'Quadratic')
QT_TRANSLATE_NOOP('Fade', 'Quadratic2')


class FadeComboBox(QComboBox):
    FadeOutIcons = {
        'Linear': QIcon.fromTheme('fadeout-linear'),
        'Quadratic': QIcon.fromTheme('fadeout-quadratic'),
        'Quadratic2': QIcon.fromTheme('fadeout-quadratic2')
    }

    FadeInIcons = {
        'Linear': QIcon.fromTheme('fadein-linear'),
        'Quadratic': QIcon.fromTheme('fadein-quadratic'),
        'Quadratic2': QIcon.fromTheme('fadein-quadratic2')
    }

    class Mode(Enum):
        FadeIn = 0
        FadeOut = 1

    def __init__(self, *args, mode=Mode.FadeOut, **kwargs):
        super().__init__(*args, **kwargs)
        self.setItemDelegate(QStyledItemDelegate())

        if mode == self.Mode.FadeIn:
            items = self.FadeInIcons
        else:
            items = self.FadeOutIcons

        for key in sorted(items.keys()):
            self.addItem(items[key], translate('Fade', key), key)

    def setCurrentType(self, type):
        self.setCurrentText(translate('Fade', type))

    def currentType(self):
        return self.currentData()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from os import cpu_count

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QDialogButtonBox, QRadioButton, QSpinBox, \
    QCheckBox, QFrame, QLabel, QGridLayout, QButtonGroup, QProgressDialog

from lisp.ui.ui_utils import translate


class GainUi(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowModality(Qt.ApplicationModal)
        self.setMaximumSize(380, 210)
        self.setMinimumSize(380, 210)
        self.resize(380, 210)

        self.setLayout(QGridLayout())

        # ReplayGain mode

        self.gainRadio = QRadioButton(self)
        self.gainRadio.setChecked(True)
        self.layout().addWidget(self.gainRadio, 0, 0)

        self.gainSpinBox = QSpinBox(self)
        self.gainSpinBox.setMaximum(150)
        self.gainSpinBox.setValue(89)
        self.layout().addWidget(self.gainSpinBox, 0, 1)

        # Normalize mode

        self.normalizeRadio = QRadioButton(self)
        self.layout().addWidget(self.normalizeRadio, 1, 0)

        self.normalizeSpinBox = QSpinBox(self)
        self.normalizeSpinBox.setRange(-100, 0)
        self.normalizeSpinBox.setEnabled(False)
        self.layout().addWidget(self.normalizeSpinBox, 1, 1)

        # All/Only selected

        self.selectionMode = QCheckBox(self)
        self.layout().addWidget(self.selectionMode, 2, 0, 1, 2)

        self.line = QFrame(self)
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)
        self.layout().addWidget(self.line, 3, 0, 1, 2)

        # Max threads (up to cpu count)

        self.cpuLabel = QLabel(self)
        self.layout().addWidget(self.cpuLabel, 4, 0)

        self.cpuSpinBox = QSpinBox(self)
        max_cpu = cpu_count()
        max_cpu = max_cpu if max_cpu is not None else 1
        self.cpuSpinBox.setRange(1, max_cpu)
        self.cpuSpinBox.setValue(max_cpu)
        self.layout().addWidget(self.cpuSpinBox, 4, 1)

        self.dialogButtons = QDialogButtonBox(self)
        self.dialogButtons.setStandardButtons(QDialogButtonBox.Ok |
                                              QDialogButtonBox.Cancel)
        self.layout().addWidget(self.dialogButtons, 5, 0, 1, 2)

        self.normalizeRadio.toggled.connect(self.normalizeSpinBox.setEnabled)
        self.normalizeRadio.toggled.connect(self.gainSpinBox.setDisabled)

        self.dialogButtons.accepted.connect(self.accept)
        self.dialogButtons.rejected.connect(self.reject)

        self.modeGroup = QButtonGroup()
        self.modeGroup.addButton(self.gainRadio)
        self.modeGroup.addButton(self.normalizeRadio)

        self.retranslateUi()

    def retranslateUi(self):
        self.setWindowTitle(
            translate('ReplayGain', 'ReplayGain / Normalization'))
        self.cpuLabel.setText(translate('ReplayGain', 'Threads number'))
        self.selectionMode.setText(
            translate('ReplayGain', 'Apply only to selected media'))
        self.gainRadio.setText(
            translate('ReplayGain', 'ReplayGain to (dB SPL)'))
        self.normalizeRadio.setText(
            translate('ReplayGain', 'Normalize to (dB)'))

    def mode(self):
        return 0 if self.gainRadio.isChecked() else 1

    def only_selected(self):
        return self.selectionMode.isChecked()

    def norm_level(self):
        return self.normalizeSpinBox.value()

    def ref_level(self):
        return self.gainSpinBox.value()

    def threads(self):
        return self.cpuSpinBox.value()


class GainProgressDialog(QProgressDialog):
    def __init__(self, maximum, parent=None):
        super().__init__(parent)

        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowTitle(translate('ReplayGain', 'Processing files ...'))
        self.setMaximumSize(320, 110)
        self.setMinimumSize(320, 110)
        self.resize(320, 110)

        self.setMaximum(maximum)
        self.setLabelText('0 / {0}'.format(maximum))

    def on_progress(self, value):
        if value == -1:
            # Hide the progress dialog
            self.setValue(self.maximum())
            self.deleteLater()
        else:
            self.setValue(self.value() + value)
            self.setLabelText('{0} / {1}'.format(self.value(), self.maximum()))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QVBoxLayout, QListWidget, QDialogButtonBox, \
    QListWidgetItem, QDialog

from lisp.core.signal import Connection
from lisp.modules.remote.discovery import Discoverer
from lisp.ui.ui_utils import translate


class PeersDiscoveryDialog(QDialog):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.setWindowModality(Qt.ApplicationModal)
        self.setLayout(QVBoxLayout())
        self.setMaximumSize(300, 200)
        self.setMinimumSize(300, 200)
        self.resize(300, 200)

        self.listWidget = QListWidget(self)
        self.listWidget.setAlternatingRowColors(True)
        self.listWidget.setSelectionMode(self.listWidget.MultiSelection)
        self.layout().addWidget(self.listWidget)

        self.dialogButton = QDialogButtonBox(self)
        self.dialogButton.setStandardButtons(self.dialogButton.Ok)
        self.dialogButton.accepted.connect(self.accept)
        self.layout().addWidget(self.dialogButton)

        self.retranslateUi()

        self._discoverer = Discoverer()
        self._discoverer.discovered.connect(self._new_peer, Connection.QtQueued)

    def retranslateUi(self):
        self.setWindowTitle(translate('Synchronizer', 'Discovering peers ...'))

    def accept(self):
        self._discoverer.stop()
        return super().accept()

    def reject(self):
        self._discoverer.stop()
        return super().reject()

    def exec_(self):
        self._discoverer.start()
        return super().exec_()

    def get_peers(self):
        return [item.address for item in self.listWidget.selectedItems()]

    def _new_peer(self, peer):
        item = QListWidgetItem(peer[1] + ' - ' + peer[0])
        item.address = peer[0]

        self.listWidget.addItem(item)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from os.path import dirname

from lisp.backend.media_element import ElementType
from lisp.core.loading import load_classes


__INPUTS = {}
__OUTPUTS = {}
__PLUGINS = {}


def load():
    for _, element_class in load_classes(__package__, dirname(__file__)):
        if element_class.ElementType == ElementType.Input:
            __INPUTS[element_class.__name__] = element_class
        elif element_class.ElementType == ElementType.Output:
            __OUTPUTS[element_class.__name__] = element_class
        elif element_class.ElementType == ElementType.Plugin:
            __PLUGINS[element_class.__name__] = element_class


# Getter functions
def inputs():
    return __INPUTS


def input_name(class_name):
    return __INPUTS[class_name].Name


def outputs():
    return __OUTPUTS


def output_name(class_name):
    return __OUTPUTS[class_name].Name


def plugins():
    return __PLUGINS


def plugin_name(class_name):
    return __PLUGINS[class_name].Name


# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtCore
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFontMetrics
from PyQt5.QtWidgets import QGroupBox, QGridLayout, QLabel, QSlider, QVBoxLayout

from lisp.modules.gst_backend.elements.equalizer10 import Equalizer10
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class Equalizer10Settings(SettingsPage):
    ELEMENT = Equalizer10
    Name = ELEMENT.Name

    FREQ = ['30Hz', '60Hz', '120Hz', '240Hz', '475Hz', '950Hz', '1900Hz',
            '3800Hz', '7525Hz', '15KHz']

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.groupBox = QGroupBox(self)
        self.groupBox.resize(self.size())
        self.groupBox.setTitle(
            translate('Equalizer10Settings', '10 Bands Equalizer (IIR)'))
        self.groupBox.setLayout(QGridLayout())
        self.groupBox.layout().setVerticalSpacing(0)
        self.layout().addWidget(self.groupBox)

        self.sliders = {}

        for n in range(10):
            label = QLabel(self.groupBox)
            label.setMinimumWidth(QFontMetrics(label.font()).width('000'))
            label.setAlignment(QtCore.Qt.AlignCenter)
            label.setNum(0)
            self.groupBox.layout().addWidget(label, 0, n)

            slider = QSlider(self.groupBox)
            slider.setRange(-24, 12)
            slider.setPageStep(1)
            slider.setValue(0)
            slider.setOrientation(QtCore.Qt.Vertical)
            slider.valueChanged.connect(label.setNum)
            self.groupBox.layout().addWidget(slider, 1, n)
            self.groupBox.layout().setAlignment(slider, QtCore.Qt.AlignHCenter)
            self.sliders['band' + str(n)] = slider

            fLabel = QLabel(self.groupBox)
            fLabel.setStyleSheet('font-size: 8pt;')
            fLabel.setAlignment(QtCore.Qt.AlignCenter)
            fLabel.setText(self.FREQ[n])
            self.groupBox.layout().addWidget(fLabel, 2, n)

    def enable_check(self, enable):
        self.groupBox.setCheckable(enable)
        self.groupBox.setChecked(False)

    def get_settings(self):
        settings = {}

        if not (self.groupBox.isCheckable() and not self.groupBox.isChecked()):
            for band in self.sliders:
                settings[band] = self.sliders[band].value()

        return settings

    def load_settings(self, settings):
        for band in self.sliders:
            self.sliders[band].setValue(settings.get(band, 0))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2017 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from lisp.core.model_adapter import ModelAdapter
from lisp.core.proxy_model import ReadOnlyProxyModel
from lisp.cues.media_cue import MediaCue


class CueListModel(ModelAdapter):

    def __init__(self, model):
        super().__init__(model)
        self.__cues = []

    def item(self, index):
        return self.__cues[index]

    def insert(self, item, index):
        item.index = index
        self.add(item)

    def pop(self, index):
        cue = self.__cues[index]
        self.model.remove(cue)
        return cue

    def move(self, old_index, new_index):
        if old_index != new_index:
            if new_index >= len(self.__cues):
                new_index = len(self.__cues) - 1

            cue = self.__cues.pop(old_index)
            self.__cues.insert(new_index, cue)

            if old_index < new_index:
                min_index = old_index
                max_index = new_index
            else:
                min_index = new_index
                max_index = old_index

            self._update_indices(min_index, max_index + 1)
            self.item_moved.emit(old_index, new_index)

    def _model_reset(self):
        self.__cues.clear()
        self.model_reset.emit()

    def _item_added(self, item):
        if not isinstance(item.index, int) or not 0 <= item.index <= len(self.__cues):
            item.index = len(self.__cues)

        self.__cues.insert(item.index, item)
        self._update_indices(item.index)

        self.item_added.emit(item)

    def _item_removed(self, item):
        self.__cues.pop(item.index)
        self._update_indices(item.index)

        self.item_removed.emit(item)

    def _update_indices(self, start, stop=-1):
        """Update the indices of cues from start to stop-1"""
        if not 0 <= stop <= len(self.__cues):
            stop = len(self.__cues)

        for index in range(start, stop):
            self.__cues[index].index = index

    def __iter__(self):
        return self.__cues.__iter__()


class RunningCueModel(ReadOnlyProxyModel):

    def __init__(self, model):
        super().__init__(model)
        self.__playing = []

    def _item_added(self, cue):
        cue.end.connect(self._remove)
        cue.started.connect(self._add)
        cue.error.connect(self._remove)
        cue.stopped.connect(self._remove)
        cue.interrupted.connect(self._remove)

    def _item_removed(self, cue):
        cue.end.disconnect(self._remove)
        cue.started.disconnect(self._add)
        cue.error.disconnect(self._remove)
        cue.stopped.disconnect(self._remove)
        cue.interrupted.disconnect(self._remove)

        self._remove(cue)

    def _model_reset(self):
        for cue in self.model:
            self._item_removed(cue)

        self.model_reset.emit()

    def _add(self, cue):
        if cue.duration > 0 and cue not in self.__playing:
            self.__playing.append(cue)
            self.item_added.emit(cue)

    def _remove(self, cue, *args):
        try:
            self.__playing.remove(cue)
            self.item_removed.emit(cue)
        except ValueError:
            pass

    def __len__(self):
        return len(self.__playing)

    def __iter__(self):
        yield from self.__playing

    def __contains__(self, item):
        return item in self.__playing



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import logging
import socket
import traceback

from PyQt5.QtWidgets import QMenu, QAction, QMessageBox

from lisp.application import Application
from lisp.core.plugin import Plugin
from lisp.core.signal import Connection
from lisp.core.util import get_lan_ip
from lisp.ui.mainwindow import MainWindow
from lisp.ui.ui_utils import translate
from .peers_dialog import PeersDialog


class Synchronizer(Plugin):
    Name = 'Synchronizer'

    def __init__(self):
        self.syncMenu = QMenu(translate('Synchronizer', 'Synchronization'))
        self.menu_action = MainWindow().menuTools.addMenu(self.syncMenu)

        self.addPeerAction = QAction(
            translate('Synchronizer', 'Manage connected peers'), MainWindow())
        self.addPeerAction.triggered.connect(self.manage_peers)
        self.syncMenu.addAction(self.addPeerAction)

        self.showIpAction = QAction(
            translate('Synchronizer', 'Show your IP'), MainWindow())
        self.showIpAction.triggered.connect(self.show_ip)
        self.syncMenu.addAction(self.showIpAction)

        self.peers = []
        self.cue_media = {}

    def init(self):
        Application().layout.cue_executed.connect(self.remote_execute,
                                                  mode=Connection.Async)

    def manage_peers(self):
        manager = PeersDialog(self.peers, parent=MainWindow())
        manager.exec_()

    def show_ip(self):
        ip = translate('Synchronizer', 'Your IP is:') + ' ' + str(get_lan_ip())
        QMessageBox.information(MainWindow(), ' ', ip)

    def reset(self):
        self.peers.clear()
        self.cue_media.clear()

    def remote_execute(self, cue):
        for peer in self.peers:
            try:
                peer['proxy'].execute(cue.index)
            except Exception:
                logging.error('REMOTE: remote control failed')
                logging.debug('REMOTE: ' + traceback.format_exc())



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy

from PyQt5.QtCore import QT_TRANSLATE_NOOP
from PyQt5.QtWidgets import QGridLayout, QListWidget, QPushButton, \
    QListWidgetItem

from lisp.modules.gst_backend.gst_pipe_edit import GstPipeEditDialog
from lisp.modules.gst_backend.settings import pages_by_element
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class GstMediaSettings(SettingsPage):
    Name = QT_TRANSLATE_NOOP('SettingsPageName', 'Media Settings')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QGridLayout())

        self._pages = []
        self._current_page = None
        self._settings = {}
        self._check = False

        self.listWidget = QListWidget(self)
        self.layout().addWidget(self.listWidget, 0, 0)

        self.pipeButton = QPushButton(
            translate('GstMediaSettings', 'Change Pipeline'), self)
        self.layout().addWidget(self.pipeButton, 1, 0)

        self.layout().setColumnStretch(0, 2)
        self.layout().setColumnStretch(1, 5)

        self.listWidget.currentItemChanged.connect(self.__change_page)
        self.pipeButton.clicked.connect(self.__edit_pipe)

    def load_settings(self, settings):
        settings = settings.get('_media_', {})
        # Create a local copy of the configuration
        self._settings = deepcopy(settings)

        # Create the widgets
        pages = pages_by_element()
        for element in settings.get('pipe', ()):
            page = pages.get(element)

            if page is not None and issubclass(page, SettingsPage):
                page = page(parent=self)
                page.load_settings(
                    settings.get('elements', {})
                        .get(element, page.ELEMENT.properties_defaults()))
                page.setVisible(False)
                self._pages.append(page)

                item = QListWidgetItem(translate('MediaElementName', page.Name))
                self.listWidget.addItem(item)

            self.listWidget.setCurrentRow(0)

    def get_settings(self):
        settings = {'elements': {}}

        for page in self._pages:
            page_settings = page.get_settings()

            if page_settings:
                settings['elements'][page.ELEMENT.__name__] = page_settings

        # The pipeline is returned only if check is disabled
        if not self._check:
            settings['pipe'] = self._settings['pipe']

        return {'_media_': settings}

    def enable_check(self, enabled):
        self._check = enabled
        for page in self._pages:
            page.enable_check(enabled)

    def __change_page(self, current, previous):
        if current is None:
            current = previous

        if self._current_page is not None:
            self.layout().removeWidget(self._current_page)
            self._current_page.hide()

        self._current_page = self._pages[self.listWidget.row(current)]
        self._current_page.show()
        self.layout().addWidget(self._current_page, 0, 1, 2, 1)

    def __edit_pipe(self):
        # Backup the settings
        self._settings = self.get_settings()['_media_']

        # Show the dialog
        dialog = GstPipeEditDialog(self._settings.get('pipe', ()), parent=self)

        if dialog.exec_() == dialog.Accepted:
            # Reset the view
            self.listWidget.clear()
            if self._current_page is not None:
                self.layout().removeWidget(self._current_page)
                self._current_page.hide()
            self._current_page = None
            self._pages.clear()

            # Reload with the new pipeline
            self._settings['pipe'] = dialog.get_pipe()

            self.load_settings({'_media_': self._settings})
            self.enable_check(self._check)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import socket
from threading import Thread

from lisp.core.configuration import config
from lisp.core.signal import Signal

IP = config['Remote']['BindIp']
PORT = int(config['Remote']['DiscoverPort'])
# To avoid conflicts with other applications
MAGIC = config['Remote']['DiscoverMagic']


class Announcer(Thread):

    def __init__(self):
        super().__init__(daemon=True)
        # Create UDP socket
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, True)
        self._socket.bind((IP, PORT))

        self._running = True

    def run(self):
        while self._running:
            data, addr = self._socket.recvfrom(1024)
            if data is not None and addr is not None:
                data = str(data, 'utf-8')
                if data == MAGIC:
                    data += socket.gethostname()
                    self._socket.sendto(bytes(data, 'utf-8'), addr)

    def stop(self):
        self._running = False
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        self._socket.close()
        self.join()


class Discoverer(Thread):

    def __init__(self):
        super().__init__()
        self.discovered = Signal()

        # Create UDP socket
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, True)

        self._running = True
        self._cache = []

    def run(self):
        self._socket.sendto(bytes(MAGIC, 'utf-8'), ('<broadcast>', PORT))

        while self._running:
            data, addr = self._socket.recvfrom(1024)
            if data is not None and addr is not None:
                data = str(data, 'utf-8')
                # Take only the IP, discard the port
                addr = addr[0]
                # Check if valid and new announcement
                if data.startswith(MAGIC) and addr not in self._cache:
                    # Get the host name
                    host = data[len(MAGIC):]
                    # Append the adders to the cache
                    self._cache.append(addr)
                    # Emit
                    self.discovered.emit((addr, host))

    def stop(self):
        self._running = False
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        self._socket.close()
        self.join()

    def get_discovered(self):
        return self._cache.copy()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum

from PyQt5.QtCore import QT_TRANSLATE_NOOP

from lisp.application import Application
from lisp.core.signal import Connection
from lisp.cues.cue import CueAction


class CueTriggers(Enum):
    Started = QT_TRANSLATE_NOOP('CueTriggers', 'Started')
    Paused = QT_TRANSLATE_NOOP('CueTriggers', 'Paused')
    Stopped = QT_TRANSLATE_NOOP('CueTriggers', 'Stopped')
    Ended = QT_TRANSLATE_NOOP('CueTriggers','Ended')


class CueHandler:
    def __init__(self, cue, triggers):
        self.triggers = triggers
        self.cue = cue

        self.cue.started.connect(self.__started, Connection.Async)
        self.cue.paused.connect(self.__paused, Connection.Async)
        self.cue.stopped.connect(self.__stopped, Connection.Async)
        self.cue.end.connect(self.__ended, Connection.Async)

    def __paused(self):
        self.__execute(CueTriggers.Paused.value)

    def __started(self):
        self.__execute(CueTriggers.Started.value)

    def __stopped(self):
        self.__execute(CueTriggers.Stopped.value)

    def __ended(self):
        self.__execute(CueTriggers.Ended.value)

    def __execute(self, trigger):
        for target_id, action in self.triggers.get(trigger, []):
            target = Application().cue_model.get(target_id)

            if target is not None:
                target.execute(CueAction(action))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os
import re
import subprocess
import sys

try:
    from os import scandir
except ImportError:
    from scandir import scandir


parser = argparse.ArgumentParser(description='i18n utility for LiSP')
parser.add_argument('locales', nargs='*',
                    help='Locales of witch generate translations')
parser.add_argument('-n', '--noobsolete', help='Discard obsolete strings',
                    action='store_true')
parser.add_argument('-q', '--qm', help='Release .qm files', action='store_true')


args = parser.parse_args()

# pylupdate command with arguments
PYLUPDATE_CMD = ['pylupdate5']
if args.noobsolete:
    PYLUPDATE_CMD.append('-noobsolete')


def existing_locales():
    for entry in scandir('lisp/i18n'):
        if entry.name.startswith('lisp_') and entry.name.endswith('.ts'):
            yield entry.name[5:-3]

# Locales of which generate translations files
LOCALES = args.locales
if not LOCALES:
    LOCALES = list(existing_locales())
    print('>>> UPDATE EXISTING:', ', '.join(LOCALES))


def search_files(root, exclude=(), extensions=()):
    exc_regex = '^(' + '|'.join(exclude) + ').*' if exclude else '$^'
    ext_regex = '.*\.(' + '|'.join(extensions) + ')$' if extensions else '.*'

    for path, directories, filenames in os.walk(root):
        if re.match(exc_regex, path):
            continue

        path = os.path.relpath(path, root)

        for filename in filenames:
            if re.match(ext_regex, filename):
                if path:
                    filename = os.path.join(path, filename)
                yield filename


def create_pro_file(root, exclude=(), extensions=('py',)):
    base_name = os.path.basename(os.path.normpath(root))
    translations = 'TRANSLATIONS = '
    for local in LOCALES:
        translations += os.path.join('i18n', base_name + '_' + local + '.ts ')

    files = 'SOURCES = ' + ' '.join(search_files(root, exclude, extensions))

    with open(os.path.join(root, base_name + '.pro'), mode='w') as pro_file:
        pro_file.write(translations)
        pro_file.write(os.linesep)
        pro_file.write(files)


def generate_for_submodules(path, qm=False):
    # Here "modules" is used generically
    modules = [entry.path for entry in scandir(path) if entry.is_dir()]
    for module in modules:
        if os.path.exists(os.path.join(module, 'i18n')):
            create_pro_file(module)
            p_file = os.path.join(module, os.path.basename(module) + '.pro')
            if qm:
                subprocess.run(['lrelease', p_file],
                               stdout=sys.stdout,
                               stderr=sys.stderr)
            else:
                subprocess.run(PYLUPDATE_CMD + [p_file],
                               stdout=sys.stdout,
                               stderr=sys.stderr)


print('>>> UPDATE TRANSLATIONS FOR APPLICATION')
create_pro_file('lisp', exclude=('lisp/modules/', 'lisp/plugins/'))
if args.qm:
    subprocess.run(['lrelease', 'lisp/lisp.pro'],
                   stdout=sys.stdout,
                   stderr=sys.stderr)
else:
    subprocess.run(PYLUPDATE_CMD + ['lisp/lisp.pro'],
                   stdout=sys.stdout,
                   stderr=sys.stderr)

print('>>> UPDATE TRANSLATIONS FOR MODULES')
generate_for_submodules('lisp/modules', qm=args.qm)

print('>>> UPDATE TRANSLATIONS FOR PLUGINS')
generate_for_submodules('lisp/plugins', qm=args.qm)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from sortedcontainers import SortedDict

from lisp.core.model import ModelException
from lisp.core.model_adapter import ModelAdapter


class CueCartModel(ModelAdapter):
    def __init__(self, model, rows, columns):
        super().__init__(model)

        self.__cues = SortedDict()
        self.__rows = rows
        self.__columns = columns

    def flat(self, index):
        """If index is multidimensional return a flatted version.

        :rtype: int
        """
        try:
            page, row, column = index
            page *= self.__rows * self.__columns
            row *= self.__columns
            return page + row + column
        except TypeError:
            return index

    def first_empty(self):
        """Return the first empty index."""
        n = -1
        for n, index in enumerate(self.__cues.keys()):
            if n != index:
                return n

        return n + 1

    def item(self, index):
        index = self.flat(index)
        try:
            return self.__cues[index]
        except KeyError:
            raise IndexError('index out of range')

    def insert(self, item, index):
        index = self.flat(index)

        if index not in self.__cues:
            item.index = index

        self.add(item)

    def pop(self, index):
        index = self.flat(index)
        try:
            cue = self.__cues[index]
        except KeyError:
            raise IndexError('index out of range')

        self.model.remove(cue)
        return cue

    def move(self, old_index, new_index):
        old_index = self.flat(old_index)
        new_index = self.flat(new_index)

        if new_index not in self.__cues:
            self.__cues[new_index] = self.__cues.pop(old_index)
            self.__cues[new_index].index = new_index
            self.item_moved.emit(old_index, new_index)
        else:
            raise ModelException('index already used {}'.format(new_index))

    def page_edges(self, page):
        start = self.flat((page, 0, 0))
        end = self.flat((page, self.__rows, self.__columns))
        return start, end

    def remove_page(self, page, lshift=True):
        start, end = self.page_edges(page)
        for index in range(start, end + 1):
            cue = self.__cues.get(index)
            if cue is not None:
                self.remove(cue)

        if lshift:
            page_size = self.__rows * self.__columns
            for index in self.__cues.irange(minimum=end + 1):
                new_index = index - page_size
                self.__cues[new_index] = self.__cues.pop(index)
                self.__cues[new_index].index = new_index

    def iter_page(self, page):
        """Iterate over the cues of the given page"""
        start, stop = self.page_edges(page)
        for index in self.__cues.irange(start, stop):
            yield self.__cues[index]

    def _item_added(self, item):
        if item.index == -1 or item.index in self.__cues:
            item.index = self.first_empty()

        self.__cues[item.index] = item
        self.item_added.emit(item)

    def _item_removed(self, item):
        self.__cues.pop(item.index)
        self.item_removed.emit(item)

    def _model_reset(self):
        self.__cues.clear()
        self.model_reset.emit()

    def __iter__(self):
        return iter(self.__cues.values())



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import time
from threading import Event

from lisp.core.signal import Signal


class RWait:
    """Provide a resumeable-wait mechanism."""

    def __init__(self):
        self._elapsed = 0
        self._start_time = 0
        self._ended = False

        self._waiting = Event()
        self._waiting.set()
        self._is_waiting = Event()
        self._is_waiting.set()

        self.start = Signal()
        self.ended = Signal()
        self.paused = Signal()
        self.stopped = Signal()

    def wait(self, timeout, lock=None):
        """Block until the timeout is elapsed or `pause` or `stop` are called.

        If the wait is paused, the next time this function is called the timeout
        will be `total_timeout - elapsed_time`.

        :param timeout: time to wait
        :param lock: lock to release before and re-acquired after the wait
        :return: True if the wait has not been interrupted by `pause` or `stop`
        :rtype: bool
        """
        if self._is_waiting.is_set():
            # Clear the events
            self._waiting.clear()
            self._is_waiting.clear()
            # Set the start-time
            self._start_time = time.monotonic() - self._elapsed

            self.start.emit()

            if lock is not None:
                lock.release()

            # Wait for the "remaining" time
            self._ended = not self._waiting.wait(timeout - self._elapsed)

            # If the wait is ended by timeout
            if self._ended:
                self._elapsed = 0
                self.ended.emit()

            # Notify that we are not waiting anymore
            self._is_waiting.set()

            if lock is not None:
                lock.acquire()

            return self._ended

        return False

    def stop(self):
        """Stop the wait."""
        if self.current_time() > 0:
            self._waiting.set()
            self._is_waiting.wait()

            # Check self._ended to ensure that the wait is not ended by timeout
            # before `self._waiting.set()` is called in this method.
            if not self._ended:
                self._elapsed = 0
                self.stopped.emit()

    def pause(self):
        """Pause the wait."""
        if not self._is_waiting.is_set():
            # Calculate elapsed time ASAP
            elapsed = time.monotonic() - self._start_time
            self._waiting.set()
            self._is_waiting.wait()

            # Check self._ended to ensure that the wait is not ended by timeout
            # before `self._waiting.set()` is called in this method.
            if not self._ended:
                self._elapsed = elapsed
                self.paused.emit()

    def current_time(self):
        """Return the currently elapsed time."""
        if self._is_waiting.is_set():
            return self._elapsed

        return time.monotonic() - self._start_time

    def is_waiting(self):
        return not self._is_waiting.is_set()

    def is_paused(self):
        return self._is_waiting.is_set() and self.current_time() > 0



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import json

from lisp.modules.uri_changer.json_utils import json_deep_search, \
    json_deep_replace


class Session:

    def __init__(self, file):
        self.file = file
        self.session = {}
        self.prefixes = set()

        self.load()

    def load(self):
        # Read the file content
        with open(self.file, mode='r', encoding='utf-8') as file:
            self.session = json.load(file)

    def analyze(self):
        self.prefixes = set()

        for uri in json_deep_search(self.session, 'uri'):
            prefix = ''

            # The uri should be like "protocol://some/url/here"
            for token in uri[uri.find('://')+4:].split('/')[:-1]:
                prefix += '/' + token
                self.prefixes.add(prefix)

        return self.prefixes

    def replace(self, old, new):
        def replace(value):
            if isinstance(value, str):
                return value.replace(old, new)
            return value

        json_deep_replace(self.session, 'uri', replace)

    def save(self, file):
        with open(file, mode='w', encoding='utf-8') as new_file:
            new_file.write(json.dumps(self.session, sort_keys=True, indent=4))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtCore
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QGroupBox, QHBoxLayout, QSlider, QLabel, QVBoxLayout

from lisp.modules.gst_backend.elements.speed import Speed
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class SpeedSettings(SettingsPage):
    ELEMENT = Speed
    Name = ELEMENT.Name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.groupBox = QGroupBox(self)
        self.groupBox.setLayout(QHBoxLayout())
        self.layout().addWidget(self.groupBox)

        self.speedSlider = QSlider(self.groupBox)
        self.speedSlider.setMinimum(1)
        self.speedSlider.setMaximum(1000)
        self.speedSlider.setPageStep(1)
        self.speedSlider.setValue(100)
        self.speedSlider.setOrientation(QtCore.Qt.Horizontal)
        self.speedSlider.setTickPosition(QSlider.TicksAbove)
        self.speedSlider.setTickInterval(10)
        self.groupBox.layout().addWidget(self.speedSlider)

        self.speedLabel = QLabel(self.groupBox)
        self.speedLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.groupBox.layout().addWidget(self.speedLabel)

        self.groupBox.layout().setStretch(0, 3)
        self.groupBox.layout().setStretch(1, 1)

        self.speedSlider.valueChanged.connect(self.speedChanged)

        self.retranslateUi()

    def retranslateUi(self):
        self.groupBox.setTitle(translate('SpeedSettings', 'Speed'))
        self.speedLabel.setText('1.0')

    def enable_check(self, enable):
        self.groupBox.setCheckable(enable)
        self.groupBox.setChecked(False)

    def get_settings(self):
        if not (self.groupBox.isCheckable() and not self.groupBox.isChecked()):
            return {'speed': self.speedSlider.value() / 100}

        return {}

    def load_settings(self, settings):
        self.speedSlider.setValue(settings.get('speed', 1) * 100)

    def speedChanged(self, value):
        self.speedLabel.setText(str(value / 100.0))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
# Copyright 2016 Thomas Achtner <info@offtools.de>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

try:
    import ola
except ImportError:
    ola = False
    raise ImportError('OLA module not found, plugin not loaded.')

if ola:
    from .timecode import Timecode



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import QT_TRANSLATE_NOOP

from lisp.backend.media_element import ElementType, MediaType
from lisp.core.signal import Signal
from lisp.modules.gst_backend.gi_repository import Gst
from lisp.modules.gst_backend.gst_element import GstMediaElement, GstProperty


class DbMeter(GstMediaElement):
    ElementType = ElementType.Plugin
    MediaType = MediaType.Audio
    Name = QT_TRANSLATE_NOOP('MediaElementName', 'dB Meter')

    interval = GstProperty('level', default=50 * Gst.MSECOND)
    peak_ttl = GstProperty('level', default=Gst.SECOND, gst_name='peak-ttl')
    peak_falloff = GstProperty('level', default=20, gst_name='peak-falloff')

    def __init__(self, pipeline):
        super().__init__()

        self.level_ready = Signal()

        self.pipeline = pipeline
        self.level = Gst.ElementFactory.make('level', None)
        self.level.set_property('post-messages', True)
        self.level.set_property('interval', 50 * Gst.MSECOND)
        self.level.set_property('peak-ttl', Gst.SECOND)
        self.level.set_property('peak-falloff', 20)
        self.audio_convert = Gst.ElementFactory.make('audioconvert', None)

        self.pipeline.add(self.level)
        self.pipeline.add(self.audio_convert)

        self.level.link(self.audio_convert)

        bus = self.pipeline.get_bus()
        bus.add_signal_watch()
        self._handler = bus.connect('message::element', self.__on_message)

    def dispose(self):
        bus = self.pipeline.get_bus()
        bus.remove_signal_watch()
        bus.disconnect(self._handler)

    def sink(self):
        return self.level

    def src(self):
        return self.audio_convert

    def __on_message(self, bus, message):
        if message.src == self.level:
            structure = message.get_structure()
            if structure is not None and structure.has_name('level'):
                self.level_ready.emit(structure.get_value('peak'),
                                      structure.get_value('rms'),
                                      structure.get_value('decay'))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import mido

from lisp.core.configuration import config


def str_msg_to_dict(str_message):
    message = mido.parse_string(str_message)
    dict_msg = {'type': message.type}

    for name in message._spec.arguments:
        dict_msg[name] = getattr(message, name)

    return dict_msg


def dict_msg_to_str(dict_message):
    msg_type = dict_message.pop('type')
    message = mido.Message(msg_type, **dict_message)

    return mido.format_as_string(message, include_time=False)


def mido_backend():
    """Return the current backend object, or None"""
    backend = None

    if hasattr(mido, 'backend'):
        backend = mido.backend

    if backend is None:
        raise RuntimeError('MIDI backend not loaded')

    return backend


def mido_port_name(port_name, mode):
    """Transform application default-port to appropriate values for mido

    :param port_name: port-name to be checked
    :param mode: port I/O mode ['I'=input, 'O'=output]
    :return: the port-name to give to mido to open the right port
    """
    port_name = port_name

    if port_name == 'AppDefault':
        if mode.upper() == 'I':
            port_name = config['MIDI']['InputDevice']
        elif mode.upper() == 'O':
            port_name = config['MIDI']['OutputDevice']

    if port_name == 'SysDefault':
        return None
    elif mode.upper() == 'I' and port_name in mido_backend().get_input_names():
        return port_name
    elif mode.upper() == 'O' and port_name in mido_backend().get_output_names():
        return port_name



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import os

from PyQt5.QtCore import QStandardPaths, Qt
from PyQt5.QtGui import QCursor
from PyQt5.QtWidgets import QFileDialog, QApplication, QMessageBox

from lisp.backend import get_backend
from lisp.application import Application
from lisp.core.module import Module
from lisp.cues.cue_factory import CueFactory
from lisp.ui.mainwindow import MainWindow
from lisp.ui.ui_utils import qfile_filters, translate


class MediaCueMenus(Module):
    """Register menus to add MediaCue to layouts"""

    def __init__(self):
        MainWindow().register_cue_menu_action(
            translate('MediaCueMenus', 'Audio cue (from file)'),
            self.add_uri_audio_media_cue,
            category='Media cues',
            shortcut='CTRL+M')

    @staticmethod
    def add_uri_audio_media_cue():
        """Add audio MediaCue(s) form user-selected files"""

        if get_backend() is None:
            QMessageBox.critical(MainWindow(), 'Error', 'Backend not loaded')
            return

        # Default path to system "music" folder
        path = QStandardPaths.writableLocation(QStandardPaths.MusicLocation)

        # Get the backend extensions and create a filter for the Qt file-dialog
        extensions = get_backend().supported_extensions()
        filters = qfile_filters(extensions, anyfile=False)
        # Display a file-dialog for the user to choose the media-files
        files, _ = QFileDialog.getOpenFileNames(MainWindow(),
                                                translate('MediaCueMenus',
                                                          'Select media files'),
                                                path, filters)

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))

        # Create media cues, and add them to the Application cue_model
        for file in files:
            cue = CueFactory.create_cue('URIAudioCue', uri='file://' + file)
            # Use the filename without extension as cue name
            cue.name = os.path.splitext(os.path.basename(file))[0]
            Application().cue_model.add(cue)

        QApplication.restoreOverrideCursor()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import mido

from lisp.core.configuration import config
from lisp.core.module import Module
from lisp.modules.midi.midi_settings import MIDISettings
from lisp.ui.settings.app_settings import AppSettings


class Midi(Module):
    """Provide MIDI I/O functionality"""

    def __init__(self):
        # Register the settings widget
        AppSettings.register_settings_widget(MIDISettings)

        bk_name = config['MIDI']['Backend']
        try:
            # Load the backend and set as current mido backend
            mido.set_backend(mido.Backend(bk_name, load=True))
        except Exception:
            raise RuntimeError('Backend loading failed: {0}'.format(bk_name))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtCore
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QGroupBox, QGridLayout, QSpinBox, QLabel, \
    QVBoxLayout

from lisp.modules.gst_backend.elements.db_meter import DbMeter
from lisp.modules.gst_backend.gi_repository import Gst
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class DbMeterSettings(SettingsPage):

    ELEMENT = DbMeter
    Name = ELEMENT.Name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.groupBox = QGroupBox(self)
        self.groupBox.setGeometry(0, 0, self.width(), 180)
        self.groupBox.setLayout(QGridLayout())
        self.layout().addWidget(self.groupBox)

        # Time (sec/100) between two levels
        self.intervalSpin = QSpinBox(self.groupBox)
        self.intervalSpin.setRange(1, 1000)
        self.intervalSpin.setMaximum(1000)
        self.intervalSpin.setValue(50)
        self.groupBox.layout().addWidget(self.intervalSpin, 0, 0)

        self.intervalLabel = QLabel(self.groupBox)
        self.intervalLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.groupBox.layout().addWidget(self.intervalLabel, 0, 1)

        # Peak ttl (sec/100)
        self.ttlSpin = QSpinBox(self.groupBox)
        self.ttlSpin.setSingleStep(10)
        self.ttlSpin.setRange(10, 10000)
        self.ttlSpin.setValue(500)
        self.groupBox.layout().addWidget(self.ttlSpin, 1, 0)

        self.ttlLabel = QLabel(self.groupBox)
        self.ttlLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.groupBox.layout().addWidget(self.ttlLabel, 1, 1)

        # Peak falloff (unit per time)
        self.falloffSpin = QSpinBox(self.groupBox)
        self.falloffSpin.setRange(1, 100)
        self.falloffSpin.setValue(20)
        self.groupBox.layout().addWidget(self.falloffSpin, 2, 0)

        self.falloffLabel = QLabel(self.groupBox)
        self.falloffLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.groupBox.layout().addWidget(self.falloffLabel, 2, 1)

        self.retranslateUi()

    def retranslateUi(self):
        self.groupBox.setTitle(translate('DbMeterSettings', 'DbMeter settings'))
        self.intervalLabel.setText(
            translate('DbMeterSettings', 'Time between levels (ms)'))
        self.ttlLabel.setText(translate('DbMeterSettings', 'Peak ttl (ms)'))
        self.falloffLabel.setText(
            translate('DbMeterSettings', 'Peak falloff (dB/sec)'))

    def load_settings(self, settings):
        self.intervalSpin.setValue(settings.get('interval', 50) / Gst.MSECOND)
        self.ttlSpin.setValue(settings.get('peak_ttl', 500) / Gst.MSECOND)
        self.falloffSpin.setValue(settings.get('peak_falloff', 20))

    def get_settings(self):
        settings = {}

        if not (self.groupBox.isCheckable() and not self.groupBox.isChecked()):
            settings['interval'] = self.intervalSpin.value() * Gst.MSECOND
            settings['peak_ttl'] = self.ttlSpin.value() * Gst.MSECOND
            settings['peak_falloff'] = self.falloffSpin.value()

        return settings



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import logging
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed as futures_completed
from math import pow
from threading import Thread, Lock

import gi

from lisp.ui.ui_utils import translate

gi.require_version('Gst', '1.0')
from gi.repository import Gst
from PyQt5.QtWidgets import QMenu, QAction, QDialog

from lisp.application import Application
from lisp.core.action import Action
from lisp.core.actions_handler import MainActionsHandler
from lisp.core.module import Module
from lisp.core.signal import Signal, Connection
from lisp.cues.media_cue import MediaCue
from lisp.ui.mainwindow import MainWindow
from .gain_ui import GainUi, GainProgressDialog


class GainAction(Action):
    __slots__ = ('__media_list', '__new_volumes', '__old_volumes')

    def __init__(self):
        self.__media_list = []
        self.__new_volumes = []
        self.__old_volumes = []

    def add_media(self, media, new_volume):
        volume = media.element('Volume')
        if volume is not None:
            self.__media_list.append(media)
            self.__new_volumes.append(new_volume)
            self.__old_volumes.append(volume.normal_volume)

    def do(self):
        for n, media in enumerate(self.__media_list):
            volume = media.element('Volume')
            if volume is not None:
                volume.normal_volume = self.__new_volumes[n]

    def undo(self):
        for n, media in enumerate(self.__media_list):
            volume = media.element('Volume')
            if volume is not None:
                volume.normal_volume = self.__old_volumes[n]

    def redo(self):
        self.do()

    def log(self):
        return 'Replay gain volume adjusted'


class GainMainThread(Thread):
    MAX_GAIN = 20  # dB

    def __init__(self, files, threads, mode, ref_level, norm_level):
        super().__init__()
        self.setDaemon(True)

        self._futures = {}
        self._running = False
        self._action = GainAction()

        # file -> media {'filename1': [media1, media2], 'filename2': [media3]}
        self.files = files
        self.threads = threads
        self.mode = mode
        self.ref_level = ref_level
        self.norm_level = norm_level

        self.on_progress = Signal()

    def stop(self):
        self._running = False
        for future in self._futures:
            self._futures[future].stop()
            future.cancel()

    def run(self):
        self._running = True

        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            for file in self.files.keys():
                gain = GstGain(file, self.ref_level)
                self._futures[executor.submit(gain.gain)] = gain

            for future in futures_completed(self._futures):
                if self._running:
                    try:
                        self._post_process(*future.result())
                    except Exception:
                        # Call with the value stored in the GstGain object
                        self._post_process(*self._futures[future].result)
                else:
                    break

        if self._running:
            MainActionsHandler.do_action(self._action)
        else:
            logging.info('REPLY-GAIN:: Stopped by user')

        self.on_progress.emit(-1)
        self.on_progress.disconnect()

    def _post_process(self, gained, gain, peak, uri):
        if gained:
            if gain > self.MAX_GAIN:
                gain = self.MAX_GAIN

            if self.mode == 0:
                volume = min(1 / peak, pow(10, gain / 20))
            else:  # (self.mode == 1)
                volume = 1 / peak * pow(10, self.norm_level / 20)

            for media in self.files[uri]:
                self._action.add_media(media, volume)

            logging.info('REPLY-GAIN:: completed ' + uri)
        else:
            logging.error('REPLY-GAIN:: failed  ' + uri)

        self.on_progress.emit(1)


class ReplayGain(Module):
    Name = 'ReplayGain'

    def __init__(self):
        self._gain_thread = None

        # Entry in mainWindow menu
        self.menu = QMenu(translate('ReplayGain',
                                    'ReplayGain / Normalization'))
        self.menu_action = MainWindow().menuTools.addMenu(self.menu)

        self.actionGain = QAction(MainWindow())
        self.actionGain.triggered.connect(self.gain)
        self.actionGain.setText(translate('ReplayGain', 'Calculate'))
        self.menu.addAction(self.actionGain)

        self.actionReset = QAction(MainWindow())
        self.actionReset.triggered.connect(self._reset_all)
        self.actionReset.setText(translate('ReplayGain', 'Reset all'))
        self.menu.addAction(self.actionReset)

        self.actionResetSelected = QAction(MainWindow())
        self.actionResetSelected.triggered.connect(self._reset_selected)
        self.actionResetSelected.setText(translate('ReplayGain',
                                                   'Reset selected'))
        self.menu.addAction(self.actionResetSelected)

    def gain(self):
        gainUi = GainUi(MainWindow())
        gainUi.exec_()

        if gainUi.result() == QDialog.Accepted:

            files = {}
            if gainUi.only_selected():
                cues = Application().layout.get_selected_cues(MediaCue)
            else:
                cues = Application().cue_model.filter(MediaCue)

            for cue in cues:
                media = cue.media
                uri = media.input_uri()
                if uri is not None:
                    if uri not in files:
                        files[uri] = [media]
                    else:
                        files[uri].append(media)

            # Gain (main) thread
            self._gain_thread = GainMainThread(files, gainUi.threads(),
                                               gainUi.mode(),
                                               gainUi.ref_level(),
                                               gainUi.norm_level())

            # Progress dialog
            self._progress = GainProgressDialog(len(files))
            self._gain_thread.on_progress.connect(self._progress.on_progress,
                                                  mode=Connection.QtQueued)

            self._progress.show()
            self._gain_thread.start()

    def stop(self):
        if self._gain_thread is not None:
            self._gain_thread.stop()

    def terminate(self):
        self.stop()
        MainWindow().menuTools.removeAction(self.menu_action)

    def _reset_all(self):
        self._reset(Application().cue_model.filter(MediaCue))

    def _reset_selected(self):
        self._reset(Application().cue_model.filter(MediaCue))

    def _reset(self, cues):
        action = GainAction()
        for cue in cues:
            action.add_media(cue.media, 1.0)
        MainActionsHandler.do_action(action)


class GstGain:
    def __init__(self, uri, ref_level):
        self.__lock = Lock()

        self.uri = uri
        self.ref_level = ref_level
        self.result = (False, 0, 0, uri)
        self.gain_pipe = None

    # Create a pipeline with a fake audio output and get the gain levels
    def gain(self):
        pipe = 'uridecodebin uri="{0}" ! audioconvert ! rganalysis \
                reference-level={1} ! fakesink'.format(self.uri, self.ref_level)
        self.gain_pipe = Gst.parse_launch(pipe)

        gain_bus = self.gain_pipe.get_bus()
        gain_bus.add_signal_watch()
        gain_bus.connect('message', self._on_message)

        logging.info('REPLY-GAIN:: started ' + str(self.uri))
        self.gain_pipe.set_state(Gst.State.PLAYING)

        # Block here until EOS
        self.__lock.acquire(False)
        self.__lock.acquire()

        # Reset the pipe
        self.gain_pipe = None

        # Return the computation result
        return self.result

    def stop(self):
        if self.gain_pipe is not None:
            self.gain_pipe.send_event(Gst.Event.new_eos())

    def _on_message(self, bus, message):
        try:
            if message.type == Gst.MessageType.EOS and self.__lock.locked():
                self.__lock.release()
            elif message.type == Gst.MessageType.TAG:
                tags = message.parse_tag()
                tag = tags.get_double(Gst.TAG_TRACK_GAIN)
                peak = tags.get_double(Gst.TAG_TRACK_PEAK)

                if tag[0] and peak[0]:
                    self.gain_pipe.set_state(Gst.State.NULL)
                    self.result = (True, tag[1], peak[1], self.uri)

                    if self.__lock.locked():
                        self.__lock.release()
            elif message.type == Gst.MessageType.ERROR:
                logging.debug('REPLY-GAIN:: ' + str(message.parse_error()))

                self.gain_pipe.set_state(Gst.State.NULL)

                if self.__lock.locked():
                    self.__lock.release()
        except Exception:
            if self.__lock.locked():
                self.__lock.release()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from abc import abstractmethod
from collections import Sized, Iterable, Container

from lisp.core.signal import Signal


class Model(Sized, Iterable, Container):
    """A model is a data container that provide signal to observe changes.

    The base model do not provide a way to change directly the data,
    this to remove useless restrictions to models that will store complex
    objects where attribute changes but not objects directly.
    Setter methods and signals can be simply implemented in subclass, where
    needed.

    A model should be as generic as possible in the way its store data, working
    like a normal data structure, for more ui-oriented tasks a ModelAdapter
    should be used.

    __iter__ must provide an iterator over the items
    __len__ must return the number of stored items
    __contains__ must return True/False if the given item is in/not in the model
    """

    def __init__(self):
        self.item_added = Signal()
        self.item_removed = Signal()
        self.model_reset = Signal()

    @abstractmethod
    def add(self, item):
        pass

    @abstractmethod
    def remove(self, item):
        pass

    @abstractmethod
    def reset(self):
        pass


class ModelException(Exception):
    """Exception for illegal operations on models"""



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import aifc
import math
import sunau
import urllib.parse
import wave

# Decibel value to be considered -inf
MIN_VOLUME_DB = -144
# Linear value of MIN_VOLUME_DB
MIN_VOLUME = 6.30957344480193e-08
# Maximum linear value for the volume, equals to 1000%
MAX_VOLUME = 10
# Decibel value of MAX_VOLUME
MAX_VOLUME_DB = 20


def db_to_linear(value, min_db_zero=True):
    """dB value to linear value conversion."""
    if min_db_zero and value <= MIN_VOLUME_DB:
        return 0

    return 10 ** (value / 20)


def linear_to_db(value):
    """Linear value to dB value conversion."""
    return 20 * math.log10(value) if value > MIN_VOLUME else MIN_VOLUME_DB


def fader_to_slider(value):
    """Inverse function of `slider_to_fader`.

    Note::
        If converting back to an integer scale use `round()` instead of `int()`
    """
    return (value / 3.16227766) ** (1 / 3.7)


def slider_to_fader(value):
    """Convert a slider linear value to a fader-like scale.

    The formula used is the following:
        (10db) * (x ^ 3.7)

    Where 10db = 3.16227766
    And 0.0 <= x <= 1.0

    :param value: The value to scale [0-1]
    :type value: float
    """
    if value > 1.0:
        value = 1.0
    elif value < 0.0:
        value = 0

    return 3.16227766 * (value ** 3.7)


def python_duration(path, sound_module):
    """Returns audio-file duration using the given standard library module."""
    duration = 0
    try:
        with sound_module.open(path, 'r') as file:
            frames = file.getnframes()
            rate = file.getframerate()
            duration = int(frames / rate * 1000)
    finally:
        return duration


def uri_duration(uri):
    """Return the audio-file duration, using the given uri"""
    protocol, path = uri.split('://')
    path = urllib.parse.unquote(path)

    if protocol == 'file':
        for mod in [wave, aifc, sunau]:
            duration = python_duration(path, mod)
            if duration > 0:
                return duration

    return 0



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QTreeWidget, QHeaderView, QVBoxLayout, \
    QDialogButtonBox, QTreeWidgetItem

from lisp.ui import elogging


class CueSelectDialog(QDialog):
    def __init__(self, cues=None, properties=('index', 'name'),
                 selection_mode=QTreeWidget.SingleSelection, **kwargs):
        super().__init__(**kwargs)

        self.setMinimumSize(600, 400)

        self._properties = list(properties)
        self._cues = {}

        self.list = QTreeWidget(self)
        self.list.setSelectionMode(selection_mode)
        self.list.setSelectionBehavior(QTreeWidget.SelectRows)
        self.list.setAlternatingRowColors(True)
        self.list.setIndentation(0)
        self.list.setHeaderLabels([prop.title() for prop in properties])
        self.list.header().setSectionResizeMode(QHeaderView.Fixed)
        self.list.header().setSectionResizeMode(1, QHeaderView.Stretch)
        self.list.header().setStretchLastSection(False)
        self.list.sortByColumn(0, Qt.AscendingOrder)
        self.list.setSortingEnabled(True)

        if cues is not None:
            self.add_cues(cues)

        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.list)

        self.buttons = QDialogButtonBox(self)
        self.buttons.addButton(QDialogButtonBox.Cancel)
        self.buttons.addButton(QDialogButtonBox.Ok)
        self.layout().addWidget(self.buttons)

        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

    def add_cue(self, cue):
        item = QTreeWidgetItem()
        item.setTextAlignment(0, Qt.AlignCenter)

        for n, prop in enumerate(self._properties):
            try:
                item.setData(n, Qt.DisplayRole, getattr(cue, prop, 'Undefined'))
            except Exception as e:
                elogging.exception('Cannot display {0} property'.format(prop), e,
                                   dialog=False)

        self._cues[cue] = item
        item.setData(0, Qt.UserRole, cue)
        self.list.addTopLevelItem(item)

    def add_cues(self, cues):
        self.list.setSortingEnabled(False)
        for cue in cues:
            self.add_cue(cue)
        self.list.setSortingEnabled(True)

    def remove_cue(self, cue):
        index = self.list.indexOfTopLevelItem(self._cues.pop(cue))
        self.list.takeTopLevelItem(index)

    def reset(self):
        self.list.clear()
        self._cues.clear()

    def selected_cues(self):
        cues = []
        for item in self.list.selectedItems():
            cues.append(item.data(0, Qt.UserRole))
        return cues

    def selected_cue(self):
        items = self.list.selectedItems()
        if items:
            return items[0].data(0, Qt.UserRole)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5 import QtCore
from PyQt5.QtCore import Qt, QT_TRANSLATE_NOOP
from PyQt5.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QGroupBox, \
    QPushButton, QDoubleSpinBox, QGridLayout

from lisp.application import Application
from lisp.backend.audio_utils import MIN_VOLUME_DB, MAX_VOLUME_DB, \
    linear_to_db, db_to_linear
from lisp.core.decorators import async
from lisp.core.fade_functions import FadeInType, FadeOutType
from lisp.core.fader import Fader
from lisp.core.has_properties import Property
from lisp.cues.cue import Cue, CueAction
from lisp.cues.media_cue import MediaCue
from lisp.ui.cuelistdialog import CueSelectDialog
from lisp.ui.settings.cue_settings import CueSettingsRegistry
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate
from lisp.ui.widgets import FadeComboBox
from lisp.ui.widgets.fade_edit import FadeEdit


class VolumeControl(Cue):
    Name = QT_TRANSLATE_NOOP('CueName', 'Volume Control')

    target_id = Property()
    fade_type = Property(default=FadeInType.Linear.name)
    volume = Property(default=.0)

    CueActions = (CueAction.Default, CueAction.Start, CueAction.Stop,
                  CueAction.Pause, CueAction.Interrupt)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = translate('CueName', self.Name)

        self.__fader = Fader(None, 'current_volume')
        self.__init_fader()

    def __init_fader(self):
        cue = Application().cue_model.get(self.target_id)

        if isinstance(cue, MediaCue):
            volume = cue.media.element('Volume')
            if volume is not None:
                if volume is not self.__fader.target:
                    self.__fader.target = volume
                return True

        return False

    def __start__(self, fade=False):
        if self.__init_fader():
            if self.__fader.is_paused():
                self.__fader.restart()
                return True

            if self.duration > 0:
                if self.__fader.target.current_volume > self.volume:
                    self.__fade(FadeOutType[self.fade_type])
                    return True
                elif self.__fader.target.current_volume < self.volume:
                    self.__fade(FadeInType[self.fade_type])
                    return True
            else:
                self.__fader.target.current_volume = self.volume

        return False

    def __stop__(self, fade=False):
        self.__fader.stop()
        return True

    def __pause__(self, fade=False):
        self.__fader.pause()
        return True

    __interrupt__ = __stop__

    @async
    def __fade(self, fade_type):
        try:
            self.__fader.prepare()
            ended = self.__fader.fade(round(self.duration / 1000, 2),
                                      self.volume,
                                      fade_type)

            if ended:
                # to avoid approximation problems
                self.__fader.target.current_volume = self.volume
                self._ended()
        except Exception as e:
            self._error(
                translate('VolumeControl', 'Error during cue execution'),
                str(e)
            )

    def current_time(self):
        return self.__fader.current_time()


class VolumeSettings(SettingsPage):
    Name = QT_TRANSLATE_NOOP('SettingsPageName', 'Volume Settings')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.__v_edit_flag = False
        self.cue_id = -1

        cues = Application().cue_model.filter(MediaCue)
        self.cueDialog = CueSelectDialog(cues=cues, parent=self)

        self.cueGroup = QGroupBox(self)
        self.cueGroup.setLayout(QVBoxLayout())
        self.layout().addWidget(self.cueGroup)

        self.cueLabel = QLabel(self.cueGroup)
        self.cueLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.cueLabel.setStyleSheet('font-weight: bold;')
        self.cueGroup.layout().addWidget(self.cueLabel)

        self.cueButton = QPushButton(self.cueGroup)
        self.cueButton.clicked.connect(self.select_cue)
        self.cueGroup.layout().addWidget(self.cueButton)

        self.volumeGroup = QGroupBox(self)
        self.volumeGroup.setLayout(QHBoxLayout())
        self.layout().addWidget(self.volumeGroup)

        self.volumeEdit = QDoubleSpinBox(self.volumeGroup)
        self.volumeEdit.setDecimals(6)
        self.volumeEdit.setMaximum(100)
        self.volumeGroup.layout().addWidget(self.volumeEdit)

        self.percentLabel = QLabel('%', self.volumeGroup)
        self.volumeGroup.layout().addWidget(self.percentLabel)

        self.volumeDbEdit = QDoubleSpinBox(self.volumeGroup)
        self.volumeDbEdit.setRange(MIN_VOLUME_DB, MAX_VOLUME_DB)
        self.volumeDbEdit.setValue(MIN_VOLUME_DB)
        self.volumeGroup.layout().addWidget(self.volumeDbEdit)

        self.dbLabel = QLabel('dB', self.volumeGroup)
        self.volumeGroup.layout().addWidget(self.dbLabel)

        self.volumeEdit.valueChanged.connect(self.__volume_change)
        self.volumeDbEdit.valueChanged.connect(self.__db_volume_change)

        # Fade
        self.fadeGroup = QGroupBox(self)
        self.fadeGroup.setLayout(QHBoxLayout())
        self.layout().addWidget(self.fadeGroup)

        self.fadeEdit = FadeEdit(self.fadeGroup)
        self.fadeGroup.layout().addWidget(self.fadeEdit)

        self.retranslateUi()

    def retranslateUi(self):
        self.cueGroup.setTitle(translate('VolumeControl', 'Cue'))
        self.cueButton.setText(translate('VolumeControl', 'Click to select'))
        self.cueLabel.setText(translate('VolumeControl', 'Not selected'))
        self.volumeGroup.setTitle(translate('VolumeControl', 'Volume to reach'))
        self.fadeGroup.setTitle(translate('VolumeControl', 'Fade'))

    def select_cue(self):
        if self.cueDialog.exec_() == self.cueDialog.Accepted:
            cue = self.cueDialog.selected_cue()

            if cue is not None:
                self.cue_id = cue.id
                self.cueLabel.setText(cue.name)

    def enable_check(self, enabled):
        self.cueGroup.setCheckable(enabled)
        self.cueGroup.setChecked(False)

        self.volumeGroup.setCheckable(enabled)
        self.volumeGroup.setChecked(False)

        self.fadeGroup.setCheckable(enabled)
        self.volumeGroup.setChecked(False)

    def get_settings(self):
        conf = {}
        checkable = self.cueGroup.isCheckable()

        if not (checkable and not self.cueGroup.isChecked()):
            conf['target_id'] = self.cue_id
        if not (checkable and not self.volumeGroup.isCheckable()):
            conf['volume'] = self.volumeEdit.value() / 100
        if not (checkable and not self.fadeGroup.isCheckable()):
            conf['duration'] = self.fadeEdit.duration() * 1000
            conf['fade_type'] = self.fadeEdit.fadeType()

        return conf

    def load_settings(self, settings):
        cue = Application().cue_model.get(settings.get('target_id', ''))
        if cue is not None:
            self.cue_id = settings['target_id']
            self.cueLabel.setText(cue.name)

        self.volumeEdit.setValue(settings.get('volume', 0) * 100)
        self.fadeEdit.setDuration(settings.get('duration', 0) / 1000)
        self.fadeEdit.setFadeType(settings.get('fade_type', ''))

    def __volume_change(self, value):
        if not self.__v_edit_flag:
            try:
                self.__v_edit_flag = True
                self.volumeDbEdit.setValue(linear_to_db(value / 100))
            finally:
                self.__v_edit_flag = False

    def __db_volume_change(self, value):
        if not self.__v_edit_flag:
            try:
                self.__v_edit_flag = True
                self.volumeEdit.setValue(db_to_linear(value) * 100)
            finally:
                self.__v_edit_flag = False


CueSettingsRegistry().add_item(VolumeSettings, VolumeControl)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import os

from lisp.core.loading import load_classes
from lisp.ui import elogging

__PLUGINS = {}


def load_plugins():
    """Load available plugins."""
    for name, plugin in load_classes(__package__, os.path.dirname(__file__)):
        try:
            __PLUGINS[name] = plugin()
            elogging.debug('PLUGINS: Loaded "{0}"'.format(name))
        except Exception as e:
            elogging.exception('PLUGINS: Failed "{0}" load'.format(name), e)


def translations():
    base_path = os.path.dirname(os.path.realpath(__file__))
    for module in next(os.walk(base_path))[1]:
        i18n_dir = os.path.join(base_path, module, 'i18n')
        if os.path.exists(i18n_dir):
            yield os.path.join(i18n_dir, module)


def init_plugins():
    """Initialize all the plugins."""
    failed = []
    for plugin in __PLUGINS:
        try:
            __PLUGINS[plugin].init()
            elogging.debug('PLUGINS: Initialized "{0}"'.format(plugin))
        except Exception as e:
            failed.append(plugin)
            elogging.exception('PLUGINS: Failed "{0}" init'.format(plugin), e)

    for plugin in failed:
        __PLUGINS.pop(plugin)


def reset_plugins():
    """Resets all the plugins."""
    for plugin in __PLUGINS:
        try:
            __PLUGINS[plugin].reset()
            elogging.debug('PLUGINS: Reset "{0}"'.format(plugin))
        except Exception as e:
            elogging.exception('PLUGINS: Failed "{0}" reset'.format(plugin), e)


def set_plugins_settings(settings):
    for plugin in __PLUGINS.values():
        if plugin.Name in settings:
            try:
                plugin.load_settings(settings[plugin.Name])
            except Exception as e:
                elogging.exception('PLUGINS: Failed "{0}" settings load'
                                   .format(plugin.Name), e)


def get_plugin_settings():
    plugins_settings = {}

    for plugin in __PLUGINS.values():
        try:
            settings = plugin.settings()
            if settings is not None and len(settings) > 0:
                plugins_settings[plugin.Name] = settings
        except Exception as e:
            elogging.exception('PLUGINS: Failed "{0}" settings retrieve'
                               .format(plugin.Name), e)

    return plugins_settings



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import math

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import QWidget, QGridLayout, QSizePolicy, qApp
from sortedcontainers import SortedDict


class PageWidget(QWidget):
    move_drop_event = pyqtSignal(object, int, int)
    copy_drop_event = pyqtSignal(object, int, int)

    DRAG_MAGIC = 'LiSP_Drag&Drop'

    def __init__(self, rows, columns, *args):
        super().__init__(*args)
        self.setAcceptDrops(True)

        self.__rows = rows
        self.__columns = columns
        self.__widgets = SortedDict()

        self.setLayout(QGridLayout())
        self.layout().setContentsMargins(4, 4, 4, 4)
        self.init_layout()

    def init_layout(self):
        for row in range(0, self.__rows):
            self.layout().setRowStretch(row, 1)
            # item = QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)
            # self.layout().addItem(item, row, 0)

        for column in range(0, self.__columns):
            self.layout().setColumnStretch(column, 1)
            # item = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)
            # self.layout().addItem(item, 0, column)

    def add_widget(self, widget, row, column):
        self._check_index(row, column)
        if (row, column) not in self.__widgets:
            widget.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
            self.__widgets[(row, column)] = widget
            self.layout().addWidget(widget, row, column)
            widget.show()
        else:
            raise IndexError('cell {} already used'.format((row, column)))

    def take_widget(self, row, column):
        self._check_index(row, column)
        if (row, column) in self.__widgets:
            widget = self.__widgets.pop((row, column))
            widget.hide()
            self.layout().removeWidget(widget)
            return widget
        else:
            raise IndexError('cell {} is empty'.format((row, column)))

    def move_widget(self, o_row, o_column, n_row, n_column):
        widget = self.take_widget(o_row, o_column)
        self.add_widget(widget, n_row, n_column)

    def widget(self, row, column):
        self._check_index(row, column)
        return self.__widgets.get((row, column))

    def index(self, widget):
        for index, f_widget in self.__widgets.items():
            if widget is f_widget:
                return index

        return -1, -1

    def widgets(self):
        return iter(self.__widgets.values())

    def reset(self):
        self.__widgets.clear()

    def dragEnterEvent(self, event):
        if event.mimeData().hasText():
            if event.mimeData().text() == PageWidget.DRAG_MAGIC:
                event.accept()
            else:
                event.ignore()
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        event.ignore()

    def dropEvent(self, event):
        row, column = self._event_index(event)
        if self.layout().itemAtPosition(row, column) is None:
            if qApp.keyboardModifiers() == Qt.ControlModifier:
                event.setDropAction(Qt.MoveAction)
                event.accept()
                self.move_drop_event.emit(event.source(), row, column)
            elif qApp.keyboardModifiers() == Qt.ShiftModifier:
                event.setDropAction(Qt.CopyAction)
                self.copy_drop_event.emit(event.source(), row, column)
                event.accept()

        event.ignore()

    def dragMoveEvent(self, event):
        row, column = self._event_index(event)
        if self.layout().itemAtPosition(row, column) is None:
            event.accept()
        else:
            event.ignore()

    def _check_index(self, row, column):
        if not isinstance(row, int):
            raise TypeError('rows index must be integers, not {}'
                            .format(row.__class__.__name__))
        if not isinstance(column, int):
            raise TypeError('columns index must be integers, not {}'
                            .format(column.__class__.__name__))

        if not 0 <= row < self.__rows or not 0 <= column < self.__columns:
            raise IndexError('index out of bound {}'.format((row, column)))

    def _event_index(self, event):
        # Margins and spacings are equals
        space = self.layout().horizontalSpacing()
        margin = self.layout().contentsMargins().right()

        r_size = (self.height() + margin * 2) // self.__rows + space
        c_size = (self.width() + margin * 2) // self.__columns + space

        row = math.ceil(event.pos().y() / r_size) - 1
        column = math.ceil(event.pos().x() / c_size) - 1

        return row, column



from lisp.layouts.cart_layout.layout import CartLayout
from lisp.layouts.list_layout.layout import ListLayout


__LAYOUTS__ = [CartLayout, ListLayout]


def get_layouts():
    return __LAYOUTS__


def get_layout(name):
    for layout in __LAYOUTS__:
        if layout.NAME == name:
            return layout



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import math

from PyQt5 import QtCore
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QGroupBox, QGridLayout, QComboBox, QDoubleSpinBox, \
    QLabel, QVBoxLayout

from lisp.modules.gst_backend.elements.audio_dynamic import AudioDynamic
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate

MIN_dB = 0.000000312  # -100dB


class AudioDynamicSettings(SettingsPage):
    ELEMENT = AudioDynamic
    Name = ELEMENT.Name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.groupBox = QGroupBox(self)
        self.groupBox.setGeometry(0, 0, self.width(), 240)
        self.groupBox.setLayout(QGridLayout())
        self.layout().addWidget(self.groupBox)

        # AudioDynamic mode
        self.modeComboBox = QComboBox(self.groupBox)
        self.modeComboBox.addItem(
            translate('AudioDynamicSettings', 'Compressor'), 'compressor')
        self.modeComboBox.addItem(
            translate('AudioDynamicSettings', 'Expander'), 'expander')
        self.groupBox.layout().addWidget(self.modeComboBox, 0, 0, 1, 1)

        self.modeLabel = QLabel(self.groupBox)
        self.modeLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.groupBox.layout().addWidget(self.modeLabel, 0, 1, 1, 1)

        # AudioDynamic characteristic
        self.chComboBox = QComboBox(self.groupBox)
        self.chComboBox.addItem(
            translate('AudioDynamicSettings', 'Soft Knee'), 'soft-knee')
        self.chComboBox.addItem(
            translate('AudioDynamicSettings', 'Hard Knee'), 'hard-knee')
        self.groupBox.layout().addWidget(self.chComboBox, 1, 0, 1, 1)

        self.chLabel = QLabel(self.groupBox)
        self.chLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.groupBox.layout().addWidget(self.chLabel, 1, 1, 1, 1)

        # AudioDynamic ratio
        self.ratioSpin = QDoubleSpinBox(self.groupBox)
        self.groupBox.layout().addWidget(self.ratioSpin, 2, 0, 1, 1)

        self.ratioLabel = QLabel(self.groupBox)
        self.ratioLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.groupBox.layout().addWidget(self.ratioLabel, 2, 1, 1, 1)

        # AudioDynamic threshold
        self.thresholdSpin = QDoubleSpinBox(self.groupBox)
        self.thresholdSpin.setMaximum(0)
        self.thresholdSpin.setMinimum(-100)
        self.thresholdSpin.setSingleStep(1)
        self.groupBox.layout().addWidget(self.thresholdSpin, 3, 0, 1, 1)

        self.thresholdLabel = QLabel(self.groupBox)
        self.thresholdLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.groupBox.layout().addWidget(self.thresholdLabel, 3, 1, 1, 1)

        self.retranslateUi()

    def retranslateUi(self):
        self.groupBox.setTitle(
            translate('AudioDynamicSettings', 'Compressor/Expander'))
        self.modeLabel.setText(translate('AudioDynamicSettings', 'Type'))
        self.chLabel.setText(translate('AudioDynamicSettings', 'Curve Shape'))
        self.ratioLabel.setText(translate('AudioDynamicSettings', 'Ratio'))
        self.thresholdLabel.setText(
            translate('AudioDynamicSettings', 'Threshold (dB)'))

    def enable_check(self, enable):
        self.groupBox.setCheckable(enable)
        self.groupBox.setChecked(False)

    def get_settings(self):
        settings = {}

        if not (self.groupBox.isCheckable() and not self.groupBox.isChecked()):
            settings['ratio'] = self.ratioSpin.value()
            settings['threshold'] = math.pow(10,
                                             self.thresholdSpin.value() / 20)

            settings['mode'] = self.modeComboBox.currentData()
            settings['characteristics'] = self.chComboBox.currentData()

        return settings

    def load_settings(self, settings):
        self.modeComboBox.setCurrentText(
            translate('AudioDynamicSettings',
                      settings.get('mode', 'compressor')))
        self.chComboBox.setCurrentText(
            translate('AudioDynamicSettings',
                      settings.get('characteristics', 'soft-knee')))

        if settings.get('threshold', 0) == 0:
            settings['threshold'] = MIN_dB

        self.thresholdSpin.setValue(20 * math.log10(settings['threshold']))
        self.ratioSpin.setValue(settings.get('ratio', 1))



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt, QT_TRANSLATE_NOOP
from PyQt5.QtWidgets import QGroupBox, QGridLayout, QTableView, QHeaderView, \
    QPushButton, QVBoxLayout

from lisp.application import Application
from lisp.plugins.controller.protocols.protocol import Protocol
from lisp.ui.qdelegates import ComboBoxDelegate, LineEditDelegate, \
    CueActionDelegate
from lisp.ui.qmodels import SimpleTableModel
from lisp.ui.settings.settings_page import CueSettingsPage
from lisp.ui.ui_utils import translate


class Keyboard(Protocol):

    def init(self):
        Application().layout.key_pressed.connect(self.__key_pressed)

    def reset(self):
        Application().layout.key_pressed.disconnect(self.__key_pressed)

    def __key_pressed(self, key_event):
        if not key_event.isAutoRepeat() and key_event.text() != '':
            self.protocol_event.emit(key_event.text())


class KeyboardSettings(CueSettingsPage):
    Name = QT_TRANSLATE_NOOP('SettingsPageName', 'Keyboard Shortcuts')

    def __init__(self, cue_class, **kwargs):
        super().__init__(cue_class, **kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setAlignment(Qt.AlignTop)

        self.keyGroup = QGroupBox(self)
        self.keyGroup.setLayout(QGridLayout())
        self.layout().addWidget(self.keyGroup)

        self.keyboardModel = SimpleTableModel([
            translate('ControllerKeySettings', 'Key'),
            translate('ControllerKeySettings', 'Action')])

        self.keyboardView = KeyboardView(cue_class, parent=self.keyGroup)
        self.keyboardView.setModel(self.keyboardModel)
        self.keyGroup.layout().addWidget(self.keyboardView, 0, 0, 1, 2)

        self.addButton = QPushButton(self.keyGroup)
        self.addButton.clicked.connect(self.__new_key)
        self.keyGroup.layout().addWidget(self.addButton, 1, 0)

        self.removeButton = QPushButton(self.keyGroup)
        self.removeButton.clicked.connect(self.__remove_key)
        self.keyGroup.layout().addWidget(self.removeButton, 1, 1)

        self.retranslateUi()

    def retranslateUi(self):
        self.keyGroup.setTitle(translate('ControllerKeySettings', 'Shortcuts'))
        self.addButton.setText(translate('ControllerSettings', 'Add'))
        self.removeButton.setText(translate('ControllerSettings', 'Remove'))

    def enable_check(self, enabled):
        self.keyGroup.setCheckable(enabled)
        self.keyGroup.setChecked(False)

    def get_settings(self):
        settings = {}

        if not (self.keyGroup.isCheckable() and not self.keyGroup.isChecked()):
            settings['keyboard'] = self.keyboardModel.rows

        return settings

    def load_settings(self, settings):
        for key, action in settings.get('keyboard', []):
            self.keyboardModel.appendRow(key, action)

    def __new_key(self):
        self.keyboardModel.appendRow('', self._cue_class.CueActions[0].name)

    def __remove_key(self):
        self.keyboardModel.removeRow(self.keyboardView.currentIndex().row())


class KeyboardView(QTableView):

    def __init__(self, cue_class, **kwargs):
        super().__init__(**kwargs)

        self.setSelectionBehavior(QTableView.SelectRows)
        self.setSelectionMode(QTableView.SingleSelection)

        self.setShowGrid(False)
        self.setAlternatingRowColors(True)

        self.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.horizontalHeader().setHighlightSections(False)

        self.verticalHeader().sectionResizeMode(QHeaderView.Fixed)
        self.verticalHeader().setDefaultSectionSize(24)
        self.verticalHeader().setHighlightSections(False)

        self.delegates = [LineEditDelegate(max_length=1),
                          CueActionDelegate(cue_class=cue_class,
                                            mode=CueActionDelegate.Mode.Name)]

        for column, delegate in enumerate(self.delegates):
            self.setItemDelegateForColumn(column, delegate)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import QT_TRANSLATE_NOOP, Qt
from PyQt5.QtWidgets import QVBoxLayout, QDialogButtonBox, QSizePolicy, \
    QHeaderView, QTableView

from lisp.application import Application
from lisp.cues.cue import CueAction
from lisp.plugins.triggers.triggers_handler import CueTriggers
from lisp.ui.cuelistdialog import CueSelectDialog
from lisp.ui.qdelegates import ComboBoxDelegate, CueActionDelegate, \
    CueSelectionDelegate
from lisp.ui.qmodels import CueClassRole, SimpleCueListModel
from lisp.ui.settings.settings_page import SettingsPage
from lisp.ui.ui_utils import translate


class TriggersSettings(SettingsPage):
    Name = QT_TRANSLATE_NOOP('SettingsPageName', 'Triggers')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setLayout(QVBoxLayout(self))
        self.layout().setAlignment(Qt.AlignTop)

        self.cue_select = CueSelectDialog(cues=Application().cue_model)

        self.triggersModel = TriggersModel()

        self.triggersView = TriggersView(self.cue_select, parent=self)
        self.triggersView.setModel(self.triggersModel)
        self.layout().addWidget(self.triggersView)

        self.dialogButtons = QDialogButtonBox(self)
        self.dialogButtons.setSizePolicy(QSizePolicy.Minimum,
                                         QSizePolicy.Minimum)
        self.layout().addWidget(self.dialogButtons)

        self.addButton = self.dialogButtons.addButton(
            translate('TriggersSettings', 'Add'), QDialogButtonBox.ActionRole)
        self.addButton.clicked.connect(self._add_trigger)

        self.delButton = self.dialogButtons.addButton(
            translate('TriggersSettings', 'Remove'),
            QDialogButtonBox.ActionRole)
        self.delButton.clicked.connect(self._remove_trigger)

    def _add_trigger(self):
        if self.cue_select.exec():
            cue = self.cue_select.selected_cue()
            if cue is not None:
                self.triggersModel.appendRow(cue.__class__,
                                             CueTriggers.Started.value,
                                             cue.id,
                                             cue.CueActions[0])

    def _remove_trigger(self):
        self.triggersModel.removeRow(self.triggersView.currentIndex().row())

    def load_settings(self, settings):
        # Remove the edited cue from the list of possible targets
        edited_cue = Application().cue_model.get(settings.get('id'))
        if edited_cue:
            self.cue_select.remove_cue(edited_cue)

        for trigger, targets in settings.get('triggers', {}).items():
            for target, action in targets:
                target = Application().cue_model.get(target)
                if target is not None:
                    self.triggersModel.appendRow(target.__class__, trigger,
                                                 target.id, CueAction(action))

    def get_settings(self):
        triggers = {}
        for trigger, target, action in self.triggersModel.rows:
            action = action.value

            if trigger not in triggers:
                triggers[trigger] = []

            # Avoid duplicate
            if (target, action) not in triggers[trigger]:
                triggers[trigger].append((target, action))

        return {'triggers': triggers}


class TriggersView(QTableView):
    def __init__(self, cue_select, **kwargs):
        super().__init__(**kwargs)

        self.setSelectionBehavior(QTableView.SelectRows)
        self.setSelectionMode(QTableView.SingleSelection)

        self.setShowGrid(False)
        self.setAlternatingRowColors(True)

        self.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.horizontalHeader().setHighlightSections(False)

        self.verticalHeader().sectionResizeMode(QHeaderView.Fixed)
        self.verticalHeader().setDefaultSectionSize(26)
        self.verticalHeader().setHighlightSections(False)

        self.delegates = [
            ComboBoxDelegate(options=[e.value for e in CueTriggers],
                             tr_context='CueTriggers'),
            CueSelectionDelegate(cue_select),
            CueActionDelegate()
        ]

        for column, delegate in enumerate(self.delegates):
            self.setItemDelegateForColumn(column, delegate)


class TriggersModel(SimpleCueListModel):
    def __init__(self):
        # NOTE: The model does fixed-indices operations based on this list
        super().__init__([translate('TriggersSettings', 'Trigger'),
                          translate('TriggersSettings', 'Cue'),
                          translate('TriggersSettings', 'Action')])

        self.rows_cc = []

    def setData(self, index, value, role=Qt.DisplayRole):
        result = super().setData(index, value, role)

        if result and role == CueClassRole:
            if self.rows[index.row()][2] not in value.CueActions:
                self.rows[index.row()][2] = value.CueActions[0]
                self.dataChanged.emit(self.index(index.row(), 2),
                                      self.index(index.row(), 2),
                                      [Qt.DisplayRole, Qt.EditRole])

        return result



from .cueaction_combobox import CueActionComboBox
from .fade_combobox import FadeComboBox
from .qclicklabel import QClickLabel
from .qclickslider import QClickSlider
from .qcolorbutton import QColorButton
from .qdbmeter import QDbMeter
from .qmessagebox import QDetailedMessageBox
from .qmutebutton import QMuteButton
from .qsteptimeedit import QStepTimeEdit
from .qstyledslider import QStyledSlider
from .qvertiacallabel import QVerticalLabel



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from ast import literal_eval

from lisp.core.action import Action
from lisp.ui.ui_utils import translate


class UpdateCueAction(Action):

    __slots__ = ('__cue', '__new', '__old')

    def __init__(self, properties, cue):
        self.__cue = cue

        self.__new = properties
        self.__old = str(cue.properties())

    def do(self):
        self.__cue.update_properties(self.__new)
        self.__new = str(self.__new)

    def undo(self):
        self.__cue.update_properties(literal_eval(self.__old))

    def redo(self):
        self.__new = literal_eval(self.__new)
        self.do()

    def log(self):
        return translate('CueActionLog',
                         'Cue settings changed: "{}"').format(self.__cue.name)


class UpdateCuesAction(Action):

    __slots__ = ('__cues', '__new', '__old')

    def __init__(self, properties, cues):
        self.__cues = cues
        self.__new = properties
        self.__old = str([cue.properties() for cue in cues])

    def do(self):
        for cue in self.__cues:
            cue.update_properties(self.__new)

        self.__new = str(self.__new)

    def undo(self):
        for cue, old in zip(self.__cues, literal_eval(self.__old)):
            cue.update_properties(old)

    def redo(self):
        self.__new = literal_eval(self.__new)
        self.do()

    def log(self):
        return translate('CueActionLog', 'Cues settings changed.')



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.


class Plugin:
    """Interface for plugins.

    The "init" and "reset" functions are called respectively after the session
    layout is created (init) and before (reset).
    """

    Name = 'Plugin'

    def init(self):
        """Initialize the session-related components of the plugin."""

    def reset(self):
        """Reset the session-related components of plugin."""

    def settings(self):
        """Return the plugin settings.

        :rtype: dict
        """
        return {}

    def load_settings(self, settings):
        """Load the plugin settings.

        :param settings: the settings to be loaded
        :type settings: dict
        """



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import QT_TRANSLATE_NOOP
from PyQt5.QtWidgets import QHBoxLayout, QTabWidget

from lisp.plugins.controller import protocols
from lisp.ui.settings.settings_page import CueSettingsPage
from lisp.ui.ui_utils import translate


class ControllerSettings(CueSettingsPage):
    Name = QT_TRANSLATE_NOOP('SettingsPageName', 'Cue Control')

    def __init__(self, cue_class, **kwargs):
        super().__init__(cue_class, **kwargs)
        self.setLayout(QHBoxLayout())

        self._pages = []

        self.tabWidget = QTabWidget(self)
        self.layout().addWidget(self.tabWidget)

        for page in protocols.ProtocolsSettingsPages:
            page_widget = page(cue_class, parent=self)

            self.tabWidget.addTab(page_widget,
                                  translate('SettingsPageName', page.Name))
            self._pages.append(page_widget)

        self.tabWidget.setCurrentIndex(0)

    def enable_check(self, enabled):
        for page in self._pages:
            page.enable_check(enabled)

    def get_settings(self):
        settings = {}
        for page in self._pages:
            settings.update(page.get_settings())

        return {'controller': settings}

    def load_settings(self, settings):
        settings = settings.get('controller', {})

        for page in self._pages:
            page.load_settings(settings)



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from lisp import backend
from lisp.backend.backend import Backend as BaseBackend
from lisp.core.decorators import memoize
from lisp.core.module import Module
from lisp.cues.media_cue import MediaCue
from lisp.modules.gst_backend.gst_utils import gst_parse_tag_list
from lisp.modules.gst_backend.gst_utils import gst_uri_metadata, gst_mime_types, \
    gst_uri_duration
from lisp.ui.settings.app_settings import AppSettings
from lisp.ui.settings.cue_settings import CueSettingsRegistry
from lisp.modules.gst_backend import elements, settings
from lisp.modules.gst_backend.gi_repository import Gst
from lisp.modules.gst_backend.gst_cue_factories import register_factories
from lisp.modules.gst_backend.gst_media_settings import GstMediaSettings
from lisp.modules.gst_backend.gst_settings import GstSettings


class GstBackend(Module, BaseBackend):
    def __init__(self):
        # Initialize GStreamer
        Gst.init(None)

        # Register GStreamer settings widgets
        AppSettings.register_settings_widget(GstSettings)
        # Add MediaCue settings widget to the CueLayout
        CueSettingsRegistry().add_item(GstMediaSettings, MediaCue)
        # Register the GstMedia cue builder
        register_factories()

        elements.load()
        settings.load()

        backend.set_backend(self)

    def uri_duration(self, uri):
        return gst_uri_duration(uri)

    def uri_tags(self, uri):
        return gst_parse_tag_list(gst_uri_metadata(uri).get_tags())

    @memoize
    def supported_extensions(self):
        extensions = {'audio': [], 'video': []}

        for gst_mime, gst_extensions in gst_mime_types():
            for mime in ['audio', 'video']:
                if gst_mime.startswith(mime):
                    extensions[mime].extend(gst_extensions)

        return extensions



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
import sys
from itertools import chain
from os import path

from PyQt5.QtCore import QTranslator, QLocale, QLibraryInfo
from PyQt5.QtGui import QFont, QIcon
from PyQt5.QtWidgets import QApplication

from lisp import modules
from lisp import plugins
from lisp.application import Application
from lisp.core.configuration import config
from lisp.ui.styles import styles


def main():
    # Parse the command-line arguments
    parser = argparse.ArgumentParser(description='Linux Show Player')
    parser.add_argument('-f', '--file', default='', nargs='?', const='',
                        help='Session file path')
    parser.add_argument('-l', '--log', choices=['debug', 'info', 'warning'],
                        default='warning', help='Log level')
    parser.add_argument('--locale', default='', help='Force specified locale')

    args = parser.parse_args()

    # Set the logging level
    if args.log == 'debug':
        log = logging.DEBUG

        # If something bad happen at low-level (e.g. segfault) print the stack
        import faulthandler
        faulthandler.enable()
    elif args.log == 'info':
        log = logging.INFO
    else:
        log = logging.WARNING

    logging.basicConfig(
        format='%(asctime)s.%(msecs)03d %(levelname)s:: %(message)s',
        datefmt='%H:%M:%S',
        level=log
    )

    # Create the QApplication
    qt_app = QApplication(sys.argv)
    qt_app.setApplicationName('Linux Show Player')
    qt_app.setQuitOnLastWindowClosed(True)

    # Force light font, for environment with "bad" QT support.
    appFont = qt_app.font()
    appFont.setWeight(QFont.Light)
    qt_app.setFont(appFont)
    # Set icons and theme from the application configuration
    QIcon.setThemeSearchPaths(styles.IconsThemePaths)
    QIcon.setThemeName(config['Theme']['icons'])
    styles.apply_style(config['Theme']['theme'])

    # Get/Set the locale
    locale = args.locale
    if locale:
        QLocale().setDefault(QLocale(locale))

    logging.info('Using {} locale'.format(QLocale().name()))

    # Main app translations
    translator = QTranslator()
    translator.load(QLocale(), 'lisp', '_',
                    path.join(path.dirname(path.realpath(__file__)), 'i18n'))

    qt_app.installTranslator(translator)
    ui_translators = [translator]

    # Qt platform translation
    translator = QTranslator()
    translator.load(QLocale(), 'qt', '_',
                    QLibraryInfo.location(QLibraryInfo.TranslationsPath))

    qt_app.installTranslator(translator)
    ui_translators.append(translator)

    # Modules and plugins translations
    for tr_file in chain(modules.translations(),
                         plugins.translations()):
        translator = QTranslator()
        translator.load(QLocale(), tr_file, '_')

        qt_app.installTranslator(translator)
        ui_translators.append(translator)

    # Create the application
    lisp_app = Application()
    # Load modules and plugins
    modules.load_modules()
    plugins.load_plugins()

    # Start/Initialize LiSP Application
    lisp_app.start(session_file=args.file)
    # Start Qt Application (block until exit)
    exit_code = qt_app.exec_()

    # Finalize the application
    lisp_app.finalize()
    # Exit
    sys.exit(exit_code)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
from collections import namedtuple

from PyQt5.QtGui import QPalette, QColor
from PyQt5.QtWidgets import QStyleFactory, qApp

try:
    from os import scandir
except ImportError:
    from scandir import scandir

StylesPath = os.path.abspath(os.path.dirname(__file__))
IconsThemePaths = [os.path.join(StylesPath, 'icons')]

LiSPStyles = {}
Style = namedtuple('Style', ['path', 'has_qss', 'has_py'])


def styles():
    """Returns all available styles (Qt and installed)."""
    return QStyleFactory.keys() + list(LiSPStyles.keys())


def scan_styles():
    """Scan for "installed" styles."""
    LiSPStyles.clear()

    for entry in scandir(StylesPath):
        if entry.is_dir():
            has_qss = os.path.exists(os.path.join(entry.path, 'style.qss'))
            has_py = os.path.exists(os.path.join(entry.path, 'style.py'))

            if has_qss or has_py:
                LiSPStyles[entry.name.title()] = Style(
                    path=entry.path,
                    has_qss=has_qss,
                    has_py=has_py
                )


def __load_qss_style(path):
    """Read and load the stylesheet file."""
    with open(path, mode='r', encoding='utf-8') as f:
        style = f.read()

    qApp.setStyleSheet(style)


def __load_py_style(module):
    """Load the python style module."""
    __import__(module, globals(), locals(), ['*'])


def apply_style(name):
    """Load a style given its name."""
    style = LiSPStyles.get(name.title())

    if isinstance(style, Style):
        if style.has_py:
            module = __package__ + '.' + os.path.basename(style.path) + '.style'
            __load_py_style(module)

        if style.has_qss:
            __load_qss_style(os.path.join(style.path, 'style.qss'))
    else:
        qApp.setStyleSheet('')
        qApp.setStyle(QStyleFactory.create(name))


# Search for styles
scan_styles()



# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2017 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QWidget, QGridLayout, QLabel, QSizePolicy, \
    QLCDNumber, QHBoxLayout
from lisp.cues.cue import CueAction

from lisp.core.configuration import config
from lisp.core.signal import Connection
from lisp.core.util import strtime
from lisp.cues.cue_time import CueTime
from lisp.cues.media_cue import MediaCue
from lisp.layouts.list_layout.control_buttons import CueControlButtons
from lisp.ui.widgets import QClickSlider, QDbMeter


def get_running_widget(cue, **kwargs):
    if isinstance(cue, MediaCue):
        return RunningMediaCueWidget(cue, **kwargs)
    else:
        return RunningCueWidget(cue, **kwargs)


class RunningCueWidget(QWidget):
    def __init__(self, cue, **kwargs):
        super().__init__(**kwargs)
        self.setGeometry(0, 0, self.parent().viewport().width(), 80)
        self.setFocusPolicy(Qt.NoFocus)
        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 1)

        self._accurate_time = False

        self.cue = cue
        self.cue_time = CueTime(cue)
        self.cue_time.notify.connect(self._time_updated, Connection.QtQueued)

        # Use this to avoid transparent background
        self.gridLayoutWidget = QWidget(self)
        self.gridLayout = QGridLayout(self.gridLayoutWidget)
        self.gridLayout.setContentsMargins(3, 3, 3, 3)
        self.gridLayout.setSpacing(2)
        self.layout().addWidget(self.gridLayoutWidget)

        self.nameLabel = QLabel(self.gridLayoutWidget)
        self.nameLabel.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.nameLabel.setText(cue.name)
        self.nameLabel.setToolTip(cue.name)
        self.gridLayout.addWidget(self.nameLabel, 0, 0, 1, 2)

        self.controlButtons = CueControlButtons(parent=self.gridLayoutWidget)
        self.gridLayout.addWidget(self.controlButtons, 1, 0)

        if CueAction.Stop in cue.CueActions:
            self.controlButtons.stopButton.clicked.connect(self._stop)
        else:
            self.controlButtons.stopButton.setEnabled(False)
        if CueAction.Pause in cue.CueActions:
            self.controlButtons.pauseButton.clicked.connect(self._pause)
            self.controlButtons.startButton.clicked.connect(self._restart)
        else:
            self.controlButtons.pauseButton.setEnabled(False)
            self.controlButtons.startButton.setEnabled(False)
        if CueAction.FadeIn in cue.CueActions:
            self.controlButtons.fadeInButton.clicked.connect(self._fadein)
        else:
            self.controlButtons.fadeInButton.setEnabled(False)
        if CueAction.FadeOut in cue.CueActions:
            self.controlButtons.fadeOutButton.clicked.connect(self._fadeout)
        else:
            self.controlButtons.fadeOutButton.setEnabled(False)
        if CueAction.Interrupt in cue.CueActions:
            self.controlButtons.interruptButton.clicked.connect(self._interrupt)
        else:
            self.controlButtons.interruptButton.setEnabled(False)

        self.timeDisplay = QLCDNumber(self.gridLayoutWidget)
        self.timeDisplay.setStyleSheet('background-color: transparent')
        self.timeDisplay.setSegmentStyle(QLCDNumber.Flat)
        self.timeDisplay.setDigitCount(8)
        self.timeDisplay.display(strtime(cue.duration))
        self.gridLayout.addWidget(self.timeDisplay, 1, 1)

        self.gridLayout.setRowStretch(0, 1)
        self.gridLayout.setRowStretch(1, 3)
        self.gridLayout.setColumnStretch(0, 7)
        self.gridLayout.setColumnStretch(1, 5)

        cue.changed('name').connect(self.name_changed, Connection.QtQueued)
        cue.started.connect(self.controlButtons.pauseMode, Connection.QtQueued)
        cue.paused.connect(self.controlButtons.startMode, Connection.QtQueued)

        # Fade enter/exit
        cue.fadein_start.connect(self.enter_fadein, Connection.QtQueued)
        cue.fadein_end.connect(self.exit_fade, Connection.QtQueued)
        cue.fadeout_start.connect(self.enter_fadeout, Connection.QtQueued)
        cue.fadeout_end.connect(self.exit_fade, Connection.QtQueued)

    def enter_fadein(self):
        p = self.timeDisplay.palette()
        p.setColor(p.Text, QColor(0, 255, 0))
        self.timeDisplay.setPalette(p)

    def enter_fadeout(self):
        p = self.timeDisplay.palette()
        p.setColor(p.Text, QColor(255, 50, 50))
        self.timeDisplay.setPalette(p)

    def exit_fade(self):
        self.timeDisplay.setPalette(self.palette())

    def name_changed(self, name):
        self.nameLabel.setText(name)
        self.nameLabel.setToolTip(name)

    def set_accurate_time(self, enable):
        self._accurate_time = enable

    def _time_updated(self, time):
        if not self.visibleRegion().isEmpty():
            # If the given value is the duration or < 0 set the time to 0
            if time == self.cue.duration or time < 0:
                time = 0

            self._update_timers(time)

    def _update_timers(self, time):
        self.timeDisplay.display(strtime(self.cue.duration - time,
                                         accurate=self._accurate_time))

    def _pause(self):
        self.cue.pause(fade=config['ListLayout'].getboolean('PauseCueFade'))

    def _restart(self):
        self.cue.restart(fade=config['ListLayout'].getboolean('RestartCueFade'))

    def _stop(self):
        self.cue.stop(fade=config['ListLayout'].getboolean('StopCueFade'))

    def _interrupt(self):
        self.cue.interrupt(
            fade=config['ListLayout'].getboolean('InterruptCueFade'))

    def _fadeout(self):
        self.cue.execute(CueAction.FadeOut)

    def _fadein(self):
        self.cue.execute(CueAction.FadeIn)


class RunningMediaCueWidget(RunningCueWidget):
    def __init__(self, cue, **kwargs):
        super().__init__(cue, **kwargs)
        self.setGeometry(0, 0, self.width(), 110)

        self._dbmeter_element = None

        self.seekSlider = QClickSlider(self.gridLayoutWidget)
        self.seekSlider.setOrientation(Qt.Horizontal)
        self.seekSlider.setRange(0, cue.duration)
        self.seekSlider.setFocusPolicy(Qt.NoFocus)
        self.seekSlider.sliderMoved.connect(self._seek)
        self.seekSlider.sliderJumped.connect(self._seek)
        self.seekSlider.setVisible(False)

        self.dbmeter = QDbMeter(self.gridLayoutWidget)
        self.dbmeter.setVisible(False)

        cue.changed('duration').connect(self._update_duration,
                                        Connection.QtQueued)

    def set_seek_visible(self, visible):
        if visible and not self.seekSlider.isVisible():
            self.gridLayout.addWidget(self.seekSlider, 2, 0, 1, 2)
            self.gridLayout.setRowStretch(2, 1)
        elif not visible and self.seekSlider.isVisible():
            self.gridLayout.removeWidget(self.seekSlider)
            self.gridLayout.setRowStretch(2, 0)

        self.seekSlider.setVisible(visible)

    def set_dbmeter_visible(self, visible):
        if self._dbmeter_element is not None:
            self._dbmeter_element.level_ready.disconnect(self.dbmeter.plot)
            self._dbmeter_element = None

        if visible:
            self._dbmeter_element = self.cue.media.element('DbMeter')
            if self._dbmeter_element is not None:
                self._dbmeter_element.level_ready.connect(self.dbmeter.plot)

        # Add/Remove the QDbMeter in the layout
        if visible and not self.dbmeter.isVisible():
            self.gridLayout.addWidget(self.dbmeter, 0, 2, 3, 1)
            self.gridLayout.setColumnStretch(2, 1)
        elif not visible and self.dbmeter.isVisible():
            self.gridLayout.removeWidget(self.dbmeter)
            self.gridLayout.setColumnStretch(2, 0)

        self.dbmeter.setVisible(visible)

    def _seek(self, position):
        self.cue.media.seek(position)

    def _update_duration(self, duration):
        self.seekSlider.setMaximum(duration)

    def _update_timers(self, time):
        super()._update_timers(time)
        self.seekSlider.setValue(time)


# -*- coding: utf-8 -*-
#
# This file is part of Linux Show Player
#
# Copyright 2012-2016 Francesco Ceruti <ceppofrancy@gmail.com>
#
# Linux Show Player is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linux Show Player is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linux Show Player.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABCMeta

from PyQt5.QtCore import QObject

from lisp.core.qmeta import QABCMeta


class Singleton(type):

    def __call__(cls, *args, **kwargs):
        try:
            return cls.__instance
        except AttributeError:
            cls.__instance = super(Singleton, cls).__call__(*args, **kwargs)
            return cls.__instance


class ABCSingleton(ABCMeta):

    def __call__(cls, *args, **kwargs):
        try:
            return cls.__instance
        except AttributeError:
            cls.__instance = super(ABCSingleton, cls).__call__(*args, **kwargs)
            return cls.__instance


class QSingleton(type(QObject)):

    def __call__(cls, *args, **kwargs):
        try:
            return cls.__instance
        except AttributeError:
            cls.__instance = super(QSingleton, cls).__call__(*args, **kwargs)
            return cls.__instance


class QABCSingleton(QABCMeta):

    def __call__(cls, *args, **kwargs):
        try:
            return cls.__instance
        except AttributeError:
            cls.__instance = super(QABCSingleton, cls).__call__(*args, **kwargs)
            return cls.__instance
